"""Components that are placed according to the `Layout` of the `Page`."""
from vizro.models._components.button import Button
from vizro.models._components.card import Card
from vizro.models._components.graph import Graph

__all__ = ["Button", "Card", "Graph"]



# ruff: noqa: F403, F405
import os
from pathlib import Path

import chromedriver_autoinstaller
import pytest

from vizro import Vizro


# Use monkeypatch as a session-scoped fixture.
# Taken from https://github.com/pytest-dev/pytest/issues/363#issuecomment-1335631998.
@pytest.fixture(scope="session")
def monkeypatch_session():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="session", autouse=True)
def setup_integration_test_environment(monkeypatch_session):
    # Dash debug mode seems to interfere with the tests, so we disable it here. Note "false" as a string is correct.
    monkeypatch_session.setenv("DASH_DEBUG", "false")
    # We only need to install chromedriver outside CI.
    if not os.getenv("CI"):
        chromedriver_autoinstaller.install()


@pytest.fixture
def default_dashboard(monkeypatch):
    monkeypatch.chdir(Path(__file__).parents[2] / "examples/default")
    monkeypatch.syspath_prepend(Path.cwd())
    from app import dashboard

    return dashboard


@pytest.fixture
def dict_dashboard(monkeypatch):
    monkeypatch.chdir(Path(__file__).parents[2] / "examples/from_dict")
    monkeypatch.syspath_prepend(Path.cwd())
    from app import dashboard

    return dashboard


@pytest.fixture
def json_dashboard(monkeypatch):
    monkeypatch.chdir(Path(__file__).parents[2] / "examples/from_json")
    monkeypatch.syspath_prepend(Path.cwd())
    from app import dashboard

    return dashboard


@pytest.fixture
def yaml_dashboard(monkeypatch):
    monkeypatch.chdir(Path(__file__).parents[2] / "examples/from_yaml")
    monkeypatch.syspath_prepend(Path.cwd())
    from app import dashboard

    return dashboard


def test_default_dashboard(dash_duo, default_dashboard):
    """Test if default example dashboard starts and has no errors in logs."""
    app = Vizro().build(default_dashboard).dash
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []


def test_dict_dashboard(dash_duo, dict_dashboard):
    """Test if dictionary example dashboard starts and has no errors in logs."""
    app = Vizro().build(dict_dashboard).dash
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []


def test_json_dashboard(dash_duo, json_dashboard):
    """Test if json example dashboard starts and has no errors in logs."""
    app = Vizro().build(json_dashboard).dash
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []


def test_yaml_dashboard(dash_duo, yaml_dashboard):
    """Test if yaml example dashboard starts and has no errors in logs."""
    app = Vizro().build(yaml_dashboard).dash
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



from typing import Any, List, Optional, Type, Union

from pydantic import BaseModel, Field, validator
from pydantic.fields import SHAPE_LIST, ModelField
from pydantic.typing import get_args
from typing_extensions import Annotated

from vizro.managers import model_manager
from vizro.models._models_utils import _log_call


class VizroBaseModel(BaseModel):
    """All models that are registered to the model manager should inherit from this class.

    Args:
        id (Optional[str]): ID to identify model. Must be unique throughout the whole dashboard. Defaults to `None`.
            When no ID is chosen, ID will be automatically generated.
    """

    id: Optional[str] = Field(
        None,
        description="ID to identify model. Must be unique throughout the whole dashboard."
        "When no ID is chosen, ID will be automatically generated.",
    )

    @validator("id", always=True)
    def set_id(cls, id) -> str:
        return id or model_manager._generate_id()

    @_log_call
    def __init__(self, **data: Any):
        """Adds this model instance to the model manager."""
        # Note this runs after the set_id validator, so self.id is available here. In pydantic v2 we should do this
        # using the new model_post_init method to avoid overriding __init__.
        super().__init__(**data)
        model_manager[self.id] = self

    @classmethod
    def add_type(cls, field_name: str, new_type: Type[Any]):
        """Adds a new type to an existing field based on a discriminated union.

        Args:
            field_name: Field that new type will be added to
            new_type: New type to add to discriminated union
        """

        def _add_to_discriminated_union(union):
            # Returning Annotated here isn't strictly necessary but feels safer because the new discriminated union
            # will then be annotated the same way as the old one.
            args = get_args(union)
            # args[0] is the original union, e.g. Union[Filter, Parameter]. args[1] is the FieldInfo annotated metadata.
            return Annotated[Union[args[0], new_type], args[1]]

        def _is_discriminated_union(field):
            # Really this should be done as follows:
            # return field.discriminator_key is not None
            # However, this does not work with Optional[DiscriminatedUnion]. See also TestOptionalDiscriminatedUnion.
            return hasattr(field.outer_type_, "__metadata__") and get_args(field.outer_type_)[1].discriminator

        field = cls.__fields__[field_name]
        sub_field = field.sub_fields[0] if field.shape == SHAPE_LIST else None  # type: ignore[index]

        if _is_discriminated_union(field):
            # Field itself is a non-optional discriminated union, e.g. selector: SelectorType or Optional[SelectorType].
            new_annotation = _add_to_discriminated_union(field.outer_type_)
        elif sub_field is not None and _is_discriminated_union(sub_field):
            # Field is a list of discriminated union e.g. components: List[ComponentType].
            new_annotation = List[_add_to_discriminated_union(sub_field.outer_type_)]  # type: ignore[misc,index]
        else:
            raise ValueError(
                f"Field '{field_name}' must be a discriminated union or list of discriminated union type. "
                "You probably do not need to call add_type to use your new type."
            )

        cls.__fields__[field_name] = ModelField(
            name=field.name,
            type_=new_annotation,
            class_validators=field.class_validators,
            model_config=field.model_config,
            default=field.default,
            default_factory=field.default_factory,
            required=field.required,
            final=field.final,
            alias=field.alias,
            field_info=field.field_info,
        )

        # We need to resolve all ForwardRefs again e.g. in the case of Page, which requires update_forward_refs in
        # vizro.models. The vm.__dict__.copy() is inspired by pydantic's own implementation of update_forward_refs and
        # effectively replaces all ForwardRefs defined in vizro.models.
        import vizro.models as vm

        cls.update_forward_refs(**vm.__dict__.copy())

    class Config:
        extra = "forbid"  # Good for spotting user typos and being strict.
        smart_union = True  # Makes unions work without unexpected coercion (will be the default in pydantic v2).
        validate_assignment = True  # Run validators when a field is assigned after model instantiation.
        copy_on_model_validation = "none"  # Don't copy sub-models. Essential for the model_manager to work correctly.



import logging
import os
from pathlib import Path
from typing import Dict, List, Tuple

import flask
from dash import Dash

from vizro._constants import STATIC_URL_PREFIX
from vizro.managers import data_manager, model_manager
from vizro.models import Dashboard

logger = logging.getLogger(__name__)


class Vizro:
    """The main class of the `vizro` package."""

    _user_assets_folder = Path.cwd() / "assets"
    _lib_assets_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), "static")

    def __init__(self):
        """Initialises Dash."""
        _js, _css = _append_styles(self._lib_assets_folder, STATIC_URL_PREFIX)
        self.dash = Dash(
            use_pages=True,
            pages_folder="",
            external_scripts=_js,
            external_stylesheets=_css,
            assets_folder=self._user_assets_folder,
        )

        @self.dash.server.route("/<url_prefix>/<path:filepath>")
        def serve_static(filepath, url_prefix=STATIC_URL_PREFIX):
            """Serve vizro static contents."""
            return flask.send_from_directory(self._lib_assets_folder, filepath)

    def build(self, dashboard: Dashboard):
        """Builds the dashboard.

        Args:
            dashboard (Dashboard): [`Dashboard`][vizro.models.Dashboard] object.

        Returns:
            Vizro: App object
        """
        # Note that model instantiation and pre_build are independent of Dash.
        self._pre_build()

        self.dash.layout = dashboard.build()

        return self

    def run(self, *args, **kwargs):  # if type annotated, mkdocstring stops seeing the class
        """Runs the dashboard.

        Args:
            args: Any args to `dash.run_server`
            kwargs: Any kwargs to `dash.run_server`
        """
        data_manager._frozen_state = True
        model_manager._frozen_state = True

        self.dash.run(*args, **kwargs)

    @staticmethod
    def _pre_build():
        """Runs pre_build method on all models in the model_manager."""
        # Note that a pre_build method can itself add a model (e.g. an Action) to the model manager, and so we need to
        # iterate through set(model_manager) rather than model_manager itself or we loop through something that
        # changes size.
        # Any models that are created during the pre-build process *will not* themselves have pre_build run on them.
        # In future may add a second pre_build loop after the first one.
        for model_id in set(model_manager):
            model = model_manager[model_id]
            if hasattr(model, "pre_build"):
                model.pre_build()


def _append_styles(walk_dir: str, url_prefix: str) -> Tuple[List[Dict[str, str]], List[str]]:
    """Append vizro css and js resources."""
    _vizro_css = []
    _vizro_js = []

    for current_dir, _, files in sorted(os.walk(walk_dir)):
        base = "" if current_dir == walk_dir else os.path.relpath(current_dir, walk_dir).replace("\\", "/")
        for f in sorted(files):
            path = os.path.join("/" + url_prefix, base, f) if base else os.path.join("/" + url_prefix, f)
            extension = os.path.splitext(f)[1]
            if extension == ".js":
                _vizro_js.append({"src": path, "type": "module"})
            elif extension == ".css":
                _vizro_css.append(path)
    return _vizro_js, _vizro_css



import logging
from typing import Any, Dict, List

from dash import Input, Output, State, callback, ctx
from pydantic import Field

import vizro.actions
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Action(VizroBaseModel):
    """Action to be inserted into `actions` of relevant component.

    Args:
        function (CapturedCallable): See [`CapturedCallable`][vizro.models.types.CapturedCallable].
        inputs (List[str]): Inputs in the form `<component_id>.<property>` passed to the action function.
            Defaults to `[]`.
        outputs (List[str]): Outputs in the form `<component_id>.<property>` changed by the action function.
            Defaults to `[]`.
    """

    function: CapturedCallable = Field(..., import_path=vizro.actions)
    inputs: List[str] = Field(
        [],
        description="Inputs in the form `<component_id>.<property>` passed to the action function.",
        regex="^[a-zA-Z0-9_]+[.][a-zA-Z_]+$",
    )
    outputs: List[str] = Field(
        [],
        description="Outputs in the form `<component_id>.<property>` changed by the action function.",
        regex="^[a-zA-Z0-9_]+[.][a-zA-Z_]+$",
    )

    @_log_call
    def build(self):
        from vizro.actions._callback_mapping._get_action_callback_mapping import _get_action_callback_mapping

        callback_inputs: Dict[str, Any] = {
            **_get_action_callback_mapping(action_id=self.id, argument="inputs"),  # type: ignore[arg-type]
            **{
                f'{input.split(".")[0]}_{input.split(".")[1]}': State(input.split(".")[0], input.split(".")[1])
                for input in self.inputs
            },
            "trigger": Input({"type": "action_trigger", "action_name": self.id}, "data"),
        }

        callback_outputs: Dict[str, Any] = {
            **_get_action_callback_mapping(action_id=self.id, argument="outputs"),  # type: ignore[arg-type]
            **{
                f'{output.split(".")[0]}_{output.split(".")[1]}': Output(
                    output.split(".")[0], output.split(".")[1], allow_duplicate=True
                )
                for output in self.outputs
            },
            "action_finished": Output("action_finished", "data", allow_duplicate=True),
        }

        logger.debug(
            f"Creating Callback mapping for Action ID {self.id} with "
            f"function name: {self.function._function.__name__}"
        )
        logger.debug("---------- INPUTS ----------")
        for name, object in callback_inputs.items():
            logger.debug(f"--> {name}: {object}")
        logger.debug("---------- OUTPUTS ---------")
        for name, object in callback_outputs.items():
            logger.debug(f"--> {name}: {object}")
        logger.debug("============================")

        @callback(output=callback_outputs, inputs=callback_inputs, prevent_initial_call=True)
        def callback_wrapper(trigger: None, **inputs: Dict[str, Any]):
            logger.debug("=============== ACTION ===============")
            logger.debug(f'Action ID: "{self.id}"')
            logger.debug(f'Action name: "{self.function._function.__name__}"')
            logger.debug(f"Action inputs: {inputs}")
            return_value = self.function(**inputs) or {}
            if isinstance(return_value, dict):
                return {"action_finished": None, **return_value}

            if not isinstance(return_value, list) and not isinstance(return_value, tuple):
                return_value = [return_value]

            # Map returned values to dictionary format where None belongs to the "action_finished" output
            return dict(zip(ctx.outputs_grouping.keys(), [None, *return_value]))

        return _get_action_callback_mapping(action_id=self.id, argument="components")  # type: ignore[arg-type]



import pytest

from vizro.actions._actions_utils import _create_target_arg_mapping, _update_nested_graph_properties


@pytest.fixture
def fake_graph_property_dict():
    return {"nodes": {"A": {"color": "blue"}, "B": {"color": "green"}}}


class TestUpdateNestedGraphProperties:
    def test_update_nested_graph_properties_single_level(self):
        graph = {"color": "blue"}
        result = _update_nested_graph_properties(graph, "color", "red")
        expected = {"color": "red"}
        assert result == expected

    def test_update_nested_graph_properties_multiple_levels(self):
        graph = {"node": {"label": "A", "color": "blue"}}
        result = _update_nested_graph_properties(graph, "node.color", 2)
        expected = {"node": {"label": "A", "color": 2}}
        assert result == expected

    def test_update_nested_graph_properties_nested_dict(self):
        graph = {"nodes": {"A": {"color": "blue"}, "B": {"color": "green"}}}
        result = _update_nested_graph_properties(graph, "nodes.A.color", "red")
        expected = {"nodes": {"A": {"color": "red"}, "B": {"color": "green"}}}
        assert result == expected

    def test_update_nested_graph_properties_invalid_key(self, fake_graph_property_dict):
        with pytest.raises(KeyError, match="C"):
            _update_nested_graph_properties(fake_graph_property_dict, "nodes.C.color", "red")

    def test_update_nested_graph_properties_invalid_type(self):
        graph = {"color": "blue"}
        with pytest.raises(TypeError, match="'str' object does not support item assignment"):
            _update_nested_graph_properties(graph, "color.value", 42)


class TestCreateTargetArgMapping:
    def test_single_string_one_component(self):
        input_strings = ["component1.argument1"]
        expected = {"component1": ["argument1"]}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_multiple_strings_different_components(self):
        input_strings = ["component1.argument1", "component2.argument2", "component1.argument3"]
        expected = {"component1": ["argument1", "argument3"], "component2": ["argument2"]}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_multiple_strings_same_component(self):
        input_strings = ["component1.argument1", "component1.argument2", "component1.argument3"]
        expected = {"component1": ["argument1", "argument2", "argument3"]}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_empty_input_list(self):
        input_strings = []
        expected = {}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_strings_without_separator(self):
        input_strings = ["component1_argument1", "component2_argument2"]
        with pytest.raises(ValueError, match="must contain a '.'"):
            _create_target_arg_mapping(input_strings)

    def test_strings_with_multiple_separators(self):
        input_strings = ["component1.argument1.extra", "component2.argument2.extra"]
        expected = {"component1": ["argument1.extra"], "component2": ["argument2.extra"]}
        assert _create_target_arg_mapping(input_strings) == expected



"""Unit tests for vizro.models.Navigation."""
import json

import plotly
import pytest
from pydantic import ValidationError

import vizro.models as vm
from vizro.models._navigation._accordion import Accordion


@pytest.mark.usefixtures("dashboard_build")
class TestNavigationInstantiation:
    """Tests navigation model instantiation."""

    def test_navigation_default(self):
        navigation = vm.Navigation(id="navigation")
        assert navigation.id == "navigation"
        assert navigation.pages is None

    def test_navigation_pages_as_list(self, pages_as_list):
        navigation = vm.Navigation(pages=pages_as_list, id="navigation")
        assert navigation.id == "navigation"
        assert navigation.pages == pages_as_list

    def test_navigation_pages_as_dict(self, pages_as_dict):
        navigation = vm.Navigation(pages=pages_as_dict, id="navigation")
        assert navigation.id == "navigation"
        assert navigation.pages == pages_as_dict

    def test_navigation_not_all_pages_included(self, dashboard_build):
        with pytest.warns(UserWarning):
            vm.Navigation(pages=["Page 1"])

    @pytest.mark.parametrize(
        "pages, expected_error",
        [
            ([], "Ensure this value has at least 1 item."),
            ([Accordion()], "2 validation errors for Navigation"),
            (["Page_1", "Page 2"], "1 validation error for Navigation"),
        ],
    )
    def test_field_invalid_pages_input(self, pages, expected_error):
        with pytest.raises(ValidationError, match=expected_error):
            vm.Navigation(pages=pages)


@pytest.mark.usefixtures("dashboard_build")
class TestNavigationBuild:
    """Tests navigation build method."""

    @pytest.mark.parametrize("pages", [["Page 1", "Page 2"], {"Page 1": ["Page 1"], "Page 2": ["Page 2"]}, None])
    def test_navigation_build(self, pages):
        navigation = vm.Navigation(pages=pages)
        navigation.pre_build()
        accordion = Accordion(pages=pages)
        navigation._selector.id = accordion.id

        result = json.loads(json.dumps(navigation.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(accordion.build(), cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



"""Fixtures to be shared across several tests."""
import dash_bootstrap_components as dbc
import pytest
from dash import html

from vizro import Vizro


@pytest.fixture
def dashboard_build(dashboard):
    return Vizro().build(dashboard)


@pytest.fixture()
def pages_as_list():
    return ["Page 1", "Page 2"]


@pytest.fixture
def pages_as_dict():
    return {"Page 1": ["Page 1"], "Page 2": ["Page 2"]}


@pytest.fixture
def accordion_from_page_as_list():
    accordion_buttons = [
        dbc.Button(
            children=["Page 1"],
            className="accordion-item-button",
            key="/",
            href="/",
        ),
        dbc.Button(
            children=["Page 2"],
            className="accordion-item-button",
            key="/page-2",
            href="/page-2",
        ),
    ]
    accordion_items = [
        dbc.AccordionItem(
            children=[*accordion_buttons],
            title="SELECT PAGE",
            class_name="accordion_item",
        )
    ]
    accordion = html.Div(
        children=[
            dbc.Accordion(
                id="accordion_list",
                children=accordion_items,
                class_name="accordion",
                persistence=True,
                persistence_type="session",
            ),
            html.Div(className="keyline"),
        ],
        className="nav_panel",
        id="accordion_list_outer",
    )
    return accordion


@pytest.fixture
def accordion_from_pages_as_dict():
    accordion_items = [
        dbc.AccordionItem(
            children=[dbc.Button(children=["Page 1"], className="accordion-item-button", key="/", href="/")],
            title="PAGE 1",
            class_name="accordion_item",
        ),
        dbc.AccordionItem(
            children=[
                dbc.Button(children=["Page 2"], className="accordion-item-button", key="/page-2", href="/page-2")
            ],
            title="PAGE 2",
            class_name="accordion_item",
        ),
    ]
    accordion = html.Div(
        children=[
            dbc.Accordion(
                id="accordion_dict",
                children=accordion_items,
                class_name="accordion",
                persistence=True,
                persistence_type="session",
            ),
            html.Div(className="keyline"),
        ],
        className="nav_panel",
        id="accordion_dict_outer",
    )
    return accordion



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Literal, Optional

import dash
import dash_bootstrap_components as dbc
import plotly.io as pio
from dash import ClientsideFunction, Input, Output, clientside_callback, html
from pydantic import Field, validator

import vizro
from vizro._constants import MODULE_PAGE_404, STATIC_URL_PREFIX
from vizro.actions._action_loop._action_loop import ActionLoop
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call

if TYPE_CHECKING:
    from vizro.models import Navigation, Page

logger = logging.getLogger(__name__)


def update_theme(on: bool):
    return "vizro_dark" if on else "vizro_light"


def create_layout_page_404():
    return html.Div(
        [
            html.Img(src=STATIC_URL_PREFIX + "/images/errors/error_404.svg"),
            html.Div(
                [
                    html.Div(
                        [
                            html.H3("This page could not be found.", className="heading-3-600"),
                            html.P("Make sure the URL you entered is correct."),
                        ],
                        className="error_text_container",
                    ),
                    dbc.Button("Take me home", href="/", className="button_primary"),
                ],
                className="error_content_container",
            ),
        ],
        className="page_error_container",
    )


class Dashboard(VizroBaseModel):
    """Vizro Dashboard to be used within [`Vizro`][vizro._vizro.Vizro.build].

    Args:
        pages (List[Page]): See [`Page`][vizro.models.Page].
        theme (Literal["vizro_dark", "vizro_light"]): Layout theme to be applied across dashboard.
            Defaults to `vizro_dark`.
        navigation (Optional[Navigation]): See [`Navigation`][vizro.models.Navigation]. Defaults to `None`.
        title (Optional[str]): Dashboard title to appear on every page on top left-side. Defaults to `None`.
    """

    pages: List[Page]
    theme: Literal["vizro_dark", "vizro_light"] = Field(
        "vizro_dark", description="Layout theme to be applied across dashboard. Defaults to `vizro_dark`"
    )
    navigation: Optional[Navigation] = None
    title: Optional[str] = Field(None, description="Dashboard title to appear on every page on top left-side.")

    @validator("pages", always=True)
    def validate_pages_empty_list(cls, pages):
        if not pages:
            raise ValueError("Ensure this value has at least 1 item.")
        return pages

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # pio is the backend global state and shouldn't be changed while
        # the app is running. This limitation leads to the case that Graphs blink
        # on page load if user previously has changed theme_selector.
        pio.templates.default = self.theme

    @_log_call
    def build(self):
        # Setting order here ensures that the pages in dash.page_registry preserves the order of the List[Page].
        # For now the homepage (path /) corresponds to self.pages[0].
        # Note redirect_from=["/"] doesn't work and so the / route must be defined separately.
        for order, page in enumerate(self.pages):
            path = page.path if order else "/"
            dash.register_page(module=page.id, name=page.title, path=path, order=order, layout=page.build)
            page.build()  # TODO: ideally remove, but necessary to register slider callbacks
        self._update_theme()
        self._create_error_page_404()

        return dbc.Container(
            id="dashboard_container",
            children=[
                html.Div(id=f"vizro_version_{vizro.__version__}"),
                *ActionLoop._create_app_callbacks(),
                dash.page_container,
            ],
            className=self.theme,
            fluid=True,
        )

    @staticmethod
    def _update_theme():
        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="update_dashboard_theme"),
            Output("dashboard_container", "className"),
            Input("theme_selector", "on"),
        )

    @staticmethod
    def _create_error_page_404():
        return dash.register_page(module=MODULE_PAGE_404, layout=create_layout_page_404())



import warnings
from functools import wraps


def _state_modifier(method):
    @wraps(method)
    def _wrapper(self, *args, **kwargs):
        if self._frozen_state:
            warnings.warn(
                f"{method.__qualname__} modifies global state. Doing this after the dashboard has been built will "
                f"not work reliably if you run the dashboard with multiple processes.",
                RuntimeWarning,
                stacklevel=2,
            )
        return method(self, *args, **kwargs)

    return _wrapper



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Optional

from pydantic import PrivateAttr, validator

from vizro.managers import model_manager
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models.types import NavigationPagesType

if TYPE_CHECKING:
    from vizro.models._navigation._accordion import Accordion


# Validator for re-use in other models to validate pages
def _validate_pages(pages):
    from vizro.models import Page

    if pages is not None and not pages:
        raise ValueError("Ensure this value has at least 1 item.")

    if pages:
        registered_pages = [page[0] for page in model_manager._items_with_type(Page)]

        if isinstance(pages, dict):
            missing_pages = [
                page
                for page in registered_pages
                if page not in {page for nav_pages in pages.values() for page in nav_pages}
            ]
            unknown_pages = [page for nav_pages in pages.values() for page in nav_pages if page not in registered_pages]
        else:
            missing_pages = [page for page in registered_pages if page not in pages]
            unknown_pages = [page for page in pages if page not in registered_pages]

        if missing_pages:
            warnings.warn(
                f"Not all registered pages used in Navigation 'pages'. Missing pages {missing_pages}!", UserWarning
            )

        if unknown_pages:
            raise ValueError(
                f"Unknown page ID or page title provided to Navigation 'pages'. " f"Unknown pages: {unknown_pages}"
            )

    return pages


class Navigation(VizroBaseModel):
    """Navigation in [`Dashboard`][vizro.models.Dashboard] to structure [`Pages`][vizro.models.Page].

    Args:
        pages (Optional[NavigationPagesType]): See [NavigationPagesType][vizro.models.types.NavigationPagesType].
            Defaults to `None`.
    """

    pages: Optional[NavigationPagesType] = None
    _selector: Accordion = PrivateAttr()

    # validators
    _validate_pages = validator("pages", allow_reuse=True, always=True)(_validate_pages)

    @_log_call
    def pre_build(self):
        self._set_selector()

    def _set_selector(self):
        from vizro.models._navigation._accordion import Accordion

        self._selector = Accordion(pages=self.pages)

    @_log_call
    def build(self):
        return self._selector.build()



"""The action loop creates all the required action callbacks and its components."""

from typing import List, Union

from dash import dcc, html

from vizro.actions._action_loop._action_loop_utils import _get_actions_on_registered_pages
from vizro.actions._action_loop._build_action_loop_callbacks import _build_action_loop_callbacks
from vizro.actions._action_loop._get_action_loop_components import _get_action_loop_components


class ActionLoop:
    @classmethod
    def _create_app_callbacks(cls) -> List[Union[dcc.Store, html.Div, dcc.Download]]:
        """Builds callbacks for the action loop and for each Action in the Dashboard and returns their components.

        Returns:
            List of required components for the action loop and for each `Action` in the `Dashboard`.
        """
        return cls._build_action_loop() + cls._build_actions_models()

    @staticmethod
    def _build_action_loop():
        """Builds callbacks for the action loop and returns required components for the action loop mechanism to work.

        Returns:
            List of required components for the action loop e.g. List[dcc.Store, html.Div].
        """
        _build_action_loop_callbacks()
        return _get_action_loop_components()

    @staticmethod
    def _build_actions_models():
        """Builds a callback for each `Action` model and returns required components for these callbacks.

        Returns:
            List of required components for each `Action` in the `Dashboard` e.g. List[dcc.Download]
        """
        actions = _get_actions_on_registered_pages()
        return [action_component for action in actions for action_component in action.build()]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal, Optional

import pandas as pd
from pandas.api.types import is_numeric_dtype
from pydantic import Field, PrivateAttr, validator

from vizro._constants import FILTER_ACTION_PREFIX
from vizro.actions import _filter
from vizro.managers import data_manager, model_manager
from vizro.models import Action, VizroBaseModel
from vizro.models._components.form import (
    Checklist,
    Dropdown,
    RadioItems,
    RangeSlider,
    Slider,
)
from vizro.models._models_utils import _log_call
from vizro.models.types import MultiValueType, SelectorType

if TYPE_CHECKING:
    from vizro.models import Page


# TODO: Add temporal when relevant component is available
SELECTOR_DEFAULTS = {"numerical": RangeSlider, "categorical": Dropdown}

# Ideally we might define these as NumericalSelectorType = Union[RangeSlider, Slider] etc., but that will not work
# with isinstance checks.
SELECTORS = {"numerical": (RangeSlider, Slider), "categorical": (Checklist, Dropdown, RadioItems)}


def _filter_between(series: pd.Series, value: List[float]) -> pd.Series:
    return series.between(value[0], value[1], inclusive="both")


def _filter_isin(series: pd.Series, value: MultiValueType) -> pd.Series:
    return series.isin(value)


def _get_component_page(component_id: str) -> Page:  # type: ignore[return]
    from vizro.models import Page

    for page_id, page in model_manager._items_with_type(Page):
        if any(control.id == component_id for control in page.controls):
            return page


class Filter(VizroBaseModel):
    """Filter the data supplied to `targets` on the [`Page`][vizro.models.Page].

    Examples:
        >>> print(repr(Filter(column="species")))

    Args:
        type (Literal["filter"]): Defaults to `"filter"`.
        column (str): Column of `DataFrame` to filter.
        targets (List[str]): Target component to be affected by filter. If none are given then target all components on
            the page that use `column`.
        selector (Optional[SelectorType]): See [SelectorType][vizro.models.types.SelectorType]. Defaults to `None`.
    """

    type: Literal["filter"] = "filter"
    column: str = Field(..., description="Column of DataFrame to filter.")
    targets: List[str] = Field(
        [],
        description="Target component to be affected by filter. "
        "If none are given then target all components on the page that use `column`.",
    )
    selector: Optional[SelectorType]
    _column_type: Optional[Literal["numerical", "categorical"]] = PrivateAttr()

    @validator("targets", each_item=True)
    def check_target_present(cls, target):
        if target not in model_manager:
            raise ValueError(f"Target {target} not found in model_manager.")
        return target

    @_log_call
    def pre_build(self):
        self._set_targets()
        self._set_column_type()
        self._set_selector()
        self._set_slider_values()
        self._set_categorical_selectors_options()
        self._set_actions()

    @_log_call
    def build(self):
        return self.selector.build()  # type: ignore[union-attr]

    def _set_targets(self):
        if not self.targets:
            for component in _get_component_page(self.id).components:  # type: ignore[arg-type]
                if data_manager._has_registered_data(component.id):
                    data_frame = data_manager._get_component_data(component.id)
                    if self.column in data_frame.columns:
                        self.targets.append(component.id)
            if not self.targets:
                raise ValueError(f"Selected column {self.column} not found in any dataframe on this page.")

    def _set_column_type(self):
        data_frame = data_manager._get_component_data(self.targets[0])
        if isinstance(data_frame[self.column], pd.PeriodDtype) or is_numeric_dtype(data_frame[self.column]):
            self._column_type = "numerical"
        else:
            self._column_type = "categorical"

    def _set_selector(self):
        if self.selector is None:
            self.selector = SELECTOR_DEFAULTS[self._column_type](title=self.column.title())  # type: ignore[index]
        elif not self.selector.title:
            self.selector.title = self.column.title()

    def _set_slider_values(self):
        self.selector: SelectorType
        if isinstance(self.selector, SELECTORS["numerical"]):
            if self._column_type != "numerical":
                raise ValueError(
                    f"Chosen selector {self.selector.type} is not compatible "
                    f"with {self._column_type} column '{self.column}'."
                )
            min_values = []
            max_values = []
            for target_id in self.targets:
                data_frame = data_manager._get_component_data(target_id)
                min_values.append(data_frame[self.column].min())
                max_values.append(data_frame[self.column].max())
            if not is_numeric_dtype(pd.Series(min_values)) or not is_numeric_dtype(pd.Series(max_values)):
                raise ValueError(
                    f"Non-numeric values detected in the shared data column '{self.column}' for targeted charts. "
                    f"Please ensure that the data column contains the same data type across all targeted charts."
                )
            if self.selector.min is None:
                self.selector.min = min(min_values)
            if self.selector.max is None:
                self.selector.max = max(max_values)

    def _set_categorical_selectors_options(self):
        self.selector: SelectorType
        if isinstance(self.selector, SELECTORS["categorical"]) and not self.selector.options:
            options = set()
            for target_id in self.targets:
                data_frame = data_manager._get_component_data(target_id)
                options |= set(data_frame[self.column])

            self.selector.options = sorted(options)

    def _set_actions(self):
        self.selector: SelectorType
        if not self.selector.actions:
            filter_function = _filter_between if isinstance(self.selector, RangeSlider) else _filter_isin
            self.selector.actions = [
                Action(
                    function=_filter(
                        filter_column=self.column,
                        targets=self.targets,
                        filter_function=filter_function,
                    ),
                    id=f"{FILTER_ACTION_PREFIX}_{self.id}",
                )
            ]



import logging
from typing import List, Literal

from dash import dcc
from plotly import graph_objects as go
from pydantic import Field, validator

import vizro.plotly.express as px
from vizro.managers import data_manager
from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


def create_empty_fig(message: str) -> go.Figure:
    """Creates empty go.Figure object with a display message."""
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=[None], y=[None], showlegend=False, hoverinfo="none"))
    fig.update_layout(
        xaxis={"visible": False},
        yaxis={"visible": False},
        annotations=[{"text": message, "showarrow": False, "font": {"size": 16}}],
    )
    return fig


class Graph(VizroBaseModel):
    """Wrapper for `dcc.Graph` to visualize charts in dashboard.

    Args:
        type (Literal["graph"]): Defaults to `"graph"`.
        figure (CapturedCallable): See [`CapturedCallable`][vizro.models.types.CapturedCallable].
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["graph"] = "graph"
    figure: CapturedCallable = Field(..., import_path=px)
    actions: List[Action] = []

    # Re-used validators
    _set_actions = _action_validator_factory("clickData")

    @validator("figure")
    def process_figure_data_frame(cls, figure, values):
        data_frame = figure["data_frame"]

        # Enable running px.scatter("iris") from the Python API and specification of "data_frame": "iris" through JSON.
        # In these cases, data already exists in the data manager and just needs to be linked to the component.
        if isinstance(data_frame, str):
            data_manager._add_component(values["id"], data_frame)
            return figure

        # Standard case for px.scatter(df: pd.DataFrame).
        # Extract dataframe from the captured function and put it into the data manager.
        dataset_name = str(id(data_frame))

        logger.debug("Adding data to data manager for Graph with id %s", values["id"])
        # If the dataset already exists in the data manager then it's not a problem, it just means that we don't need
        # to duplicate it. Just log the exception for debugging purposes.
        try:
            data_manager[dataset_name] = data_frame
        except ValueError as exc:
            logger.debug(exc)

        data_manager._add_component(values["id"], dataset_name)

        # No need to keep the data in the captured function any more so remove it to save memory.
        del figure["data_frame"]
        return figure

    # Convenience wrapper/syntactic sugar.
    def __call__(self, **kwargs):
        kwargs.setdefault("data_frame", data_manager._get_component_data(self.id))  # type: ignore[arg-type]
        fig = self.figure(**kwargs)

        # Remove top margin if title is provided
        if fig.layout.title.text is None:
            fig.update_layout(margin_t=24)
        return fig

    # Convenience wrapper/syntactic sugar.
    def __getitem__(self, arg_name: str):
        # pydantic discriminated union validation seems to try Graph["type"], which throws an error unless we
        # explicitly redirect it to the correct attribute.
        if arg_name == "type":
            return self.type
        return self.figure[arg_name]

    @_log_call
    def build(self):
        return dcc.Loading(
            dcc.Graph(
                id=self.id,
                # We don't do self.__call__() until the Graph is actually built. This ensures that lazy data is not
                # loaded until the graph is first shown on the screen. At the moment, we eagerly run page.build() for
                # all pages in Dashboard.build in order to register all the callbacks in advance. In future this should
                # no longer be the case so that we achieve true lazy loading.
                figure=create_empty_fig(""),
                config={
                    "autosizable": True,
                    "frameMargins": 0,
                    "responsive": True,
                },
                className="chart_container",
            ),
            color="grey",
            parent_className="chart_container",
        )



from typing import List, Literal, Optional

from dash import dcc, html
from pydantic import Field, root_validator, validator

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components.form._form_utils import get_options_and_default, validate_options_dict, validate_value
from vizro.models._models_utils import _log_call
from vizro.models.types import MultiValueType, OptionsType


class Checklist(VizroBaseModel):
    """Categorical multi-selector `Checklist` to be provided to [`Filter`][vizro.models.Filter].

    Args:
        type (Literal["checklist"]): Defaults to `"checklist"`.
        options (OptionsType): See [`OptionsType`][vizro.models.types.OptionsType]. Defaults to `[]`.
        value (Optional[MultiValueType]): See [`MultiValueType`][vizro.models.types.MultiValueType]. Defaults to `None`.
        title (Optional[str]): Title to be displayed. Defaults to `None`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["checklist"] = "checklist"
    options: OptionsType = []  # type: ignore[assignment]
    value: Optional[MultiValueType] = None
    title: Optional[str] = Field(None, description="Title to be displayed")
    actions: List[Action] = []

    # Re-used validators
    _set_actions = _action_validator_factory("value")
    _validate_options = root_validator(allow_reuse=True, pre=True)(validate_options_dict)
    _validate_value = validator("value", allow_reuse=True, always=True)(validate_value)

    @_log_call
    def build(self):
        full_options, default_value = get_options_and_default(options=self.options, multi=True)

        return html.Div(
            [
                html.P(self.title) if self.title else None,
                dcc.Checklist(
                    id=self.id,
                    options=full_options,
                    value=self.value if self.value is not None else [default_value],
                    persistence=True,
                    className="selector_body_checklist",
                ),
            ],
            className="selector_container",
            id=f"{self.id}_outer",
        )



"""Unit tests for hyphen.models.slider."""
import json

import plotly
import pytest
from dash import dcc, html
from pydantic import ValidationError

import vizro.models as vm


@pytest.fixture()
def expected_range_slider_default():
    return html.Div(
        [
            dcc.Store(
                "range_slider_callback_data",
                data={
                    "id": "range_slider",
                    "min": None,
                    "max": None,
                },
            ),
            None,
            html.Div(
                [
                    dcc.RangeSlider(
                        id="range_slider",
                        className="range_slider_control_no_space",
                        persistence=True,
                        min=None,
                        max=None,
                        marks=None,
                        value=[None, None],
                        step=None,
                    ),
                    html.Div(
                        [
                            dcc.Input(
                                id="range_slider_start_value",
                                type="number",
                                placeholder="start",
                                className="slider_input_field_no_space_left",
                                size="24px",
                                persistence=True,
                                min=None,
                                max=None,
                                value=None,
                            ),
                            dcc.Input(
                                id="range_slider_end_value",
                                type="number",
                                placeholder="end",
                                className="slider_input_field_no_space_right",
                                persistence=True,
                                min=None,
                                max=None,
                                value=None,
                            ),
                            dcc.Store(id="temp-store-range_slider-range_slider", storage_type="local"),
                        ],
                        className="slider_input_container",
                    ),
                ],
                className="range_slider_inner_container",
            ),
        ],
        className="selector_container",
        id="range_slider_outer",
    )


@pytest.fixture()
def expected_range_slider_with_optional():
    return html.Div(
        [
            dcc.Store(
                "range_slider_with_all_callback_data",
                data={
                    "id": "range_slider_with_all",
                    "min": 0,
                    "max": 10,
                },
            ),
            html.P("Title"),
            html.Div(
                [
                    dcc.RangeSlider(
                        id="range_slider_with_all",
                        min=0,
                        max=10,
                        step=1,
                        marks={},
                        className="range_slider_control",
                        value=[0, 10],
                        persistence=True,
                    ),
                    html.Div(
                        [
                            dcc.Input(
                                id="range_slider_with_all_start_value",
                                type="number",
                                placeholder="start",
                                min=0,
                                max=10,
                                className="slider_input_field_left",
                                value=0,
                                size="24px",
                                persistence=True,
                            ),
                            dcc.Input(
                                id="range_slider_with_all_end_value",
                                type="number",
                                placeholder="end",
                                min=0,
                                max=10,
                                className="slider_input_field_right",
                                value=10,
                                persistence=True,
                            ),
                            dcc.Store(id="temp-store-range_slider-range_slider_with_all", storage_type="local"),
                        ],
                        className="slider_input_container",
                    ),
                ],
                className="range_slider_inner_container",
            ),
        ],
        className="selector_container",
        id="range_slider_with_all_outer",
    )


class TestRangeSliderInstantiation:
    """Tests model instantiation."""

    def test_create_range_slider_mandatory_only(self):
        range_slider = vm.RangeSlider()

        assert hasattr(range_slider, "id")
        assert range_slider.type == "range_slider"
        assert range_slider.min is None
        assert range_slider.max is None
        assert range_slider.step is None
        assert range_slider.marks is None
        assert range_slider.value is None
        assert range_slider.title is None
        assert range_slider.actions == []

    def test_create_range_slider_mandatory_and_optional(self):
        range_slider = vm.RangeSlider(
            min=0, max=10, step=1, marks={}, value=[1, 9], title="Test title", id="range_slider_id"
        )

        assert range_slider.min == 0
        assert range_slider.max == 10
        assert range_slider.step == 1
        assert range_slider.marks == {}
        assert range_slider.value == [1, 9]
        assert range_slider.title == "Test title"
        assert range_slider.id == "range_slider_id"
        assert range_slider.actions == []
        assert range_slider.type == "range_slider"

    @pytest.mark.parametrize(
        "min, max, expected_min, expected_max",
        [(0, None, 0, None), (None, 10, None, 10), (0, 10, 0, 10), ("1", "10", 1, 10)],
    )
    def test_valid_min_max(self, min, max, expected_min, expected_max):
        range_slider = vm.RangeSlider(min=min, max=max)

        assert range_slider.min == expected_min
        assert range_slider.max == expected_max

    def test_validate_max_invalid(self):
        with pytest.raises(
            ValidationError, match="Maximum value of slider is required to be larger than minimum value."
        ):
            vm.RangeSlider(min=10, max=0)

    @pytest.mark.parametrize(
        "value, expected",
        [
            (None, None),
            ([1, 2], [1, 2]),
            ([0.1, 1.1], [0.1, 1.1]),
            ([-10, 10], [-10, 10]),
            ([10, 10], [10, 10]),
            (["1", "10"], [1, 10]),
        ],
    )
    def test_validate_slider_value_valid(self, value, expected):
        range_slider = vm.RangeSlider(min=-10, max=10, value=value)

        assert range_slider.value == expected

    @pytest.mark.parametrize(
        "value, match",
        [
            ([0], "ensure this value has at least 2 items"),
            ([], "ensure this value has at least 2 items"),
            (2, "value is not a valid list"),
            ([0, None], "1 validation error for RangeSlider"),
            ([None, None], "2 validation errors for RangeSlider"),
            ([-1, 11], "Please provide a valid value between the min and max value."),
            ([1, 2, 3], "ensure this value has at most 2 items"),
        ],
    )
    def test_validate_slider_value_invalid(self, value, match):
        with pytest.raises(ValidationError, match=match):
            vm.RangeSlider(min=0, max=10, value=value)

    @pytest.mark.parametrize(
        "step, expected",
        [(1, 1), (2.5, 2.5), (10, 10), (None, None), ("1", 1.0)],
    )
    def test_validate_step_valid(self, step, expected):
        range_slider = vm.RangeSlider(min=0, max=10, step=step)

        assert range_slider.step == expected

    def test_validate_step_invalid(self):
        with pytest.raises(
            ValidationError,
            match="The step value of the slider must be less than or equal to the difference between max and min.",
        ):
            vm.RangeSlider(min=0, max=10, step=11)

    @pytest.mark.parametrize(
        "marks, step, expected",
        [
            ({2: "2", 4: "4", 6: "6"}, 1, {}),
            ({2: "2", 4: "4", 6: "6"}, None, {2: "2", 4: "4", 6: "6"}),
            ({}, 1, {}),
        ],
    )
    def test_step_precedence_over_marks(self, marks, step, expected):
        slider = vm.RangeSlider(min=0, max=10, marks=marks, step=step)

        assert slider.marks == expected
        assert slider.step == step

    @pytest.mark.parametrize(
        "marks, expected",
        [
            ({i: str(i) for i in range(0, 10, 5)}, {i: str(i) for i in range(0, 10, 5)}),
            ({15: 15, 25: 25}, {15.0: "15", 25.0: "25"}),
            ({"15": 15, "25": 25}, {15.0: "15", 25.0: "25"}),
            (None, None),
        ],
    )
    def test_valid_marks(self, marks, expected):
        range_slider = vm.RangeSlider(min=0, max=10, marks=marks)

        assert range_slider.marks == expected

    def test_invalid_marks(self):
        with pytest.raises(
            ValidationError,
            match="2 validation errors for RangeSlider",
        ):
            vm.RangeSlider(min=1, max=10, marks={"start": 0, "end": 10})

    @pytest.mark.parametrize("step, expected", [(1, {}), (None, None)])
    def test_set_default_marks(self, step, expected):
        slider = vm.RangeSlider(min=0, max=10, step=step)
        assert slider.marks == expected

    @pytest.mark.parametrize(
        "title",
        [
            "test",
            1,
            1.0,
            """## Test header""",
            "",
        ],
    )
    def test_valid_title(self, title):
        slider = vm.RangeSlider(title=title)

        assert slider.title == str(title)

    def test_set_action_via_validator(self, test_action_function):
        range_slider = vm.RangeSlider(actions=[vm.Action(function=test_action_function)])
        actions_chain = range_slider.actions[0]

        assert actions_chain.trigger.component_property == "value"


class TestRangeSliderBuild:
    """Tests model build method."""

    def test_range_slider_build_default(self, expected_range_slider_default):
        range_slider = vm.RangeSlider(id="range_slider")
        component = range_slider.build()

        result = json.loads(json.dumps(component, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_range_slider_default, cls=plotly.utils.PlotlyJSONEncoder))

        assert result == expected

    def test_range_slider_build_with_optional(self, expected_range_slider_with_optional):
        range_slider = vm.RangeSlider(min=0, max=10, step=1, value=[0, 10], id="range_slider_with_all", title="Title")
        component = range_slider.build()

        result = json.loads(json.dumps(component, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_range_slider_with_optional, cls=plotly.utils.PlotlyJSONEncoder))

        assert result == expected



"""Fixtures to be shared across several tests."""
import pytest

import vizro.models as vm
import vizro.plotly.express as px


@pytest.fixture
def gapminder():
    return px.data.gapminder()


@pytest.fixture
def standard_px_chart(gapminder):
    return px.scatter(
        data_frame=gapminder,
        x="gdpPercap",
        y="lifeExp",
        size="pop",
        color="continent",
        hover_name="country",
        size_max=60,
    )


@pytest.fixture()
def two_pages():
    two_pages = [
        vm.Page(title="Page 1", components=[vm.Button(), vm.Button()]),
        vm.Page(title="Page 2", components=[vm.Button(), vm.Button()]),
    ]
    return two_pages


@pytest.fixture()
def dashboard(two_pages):
    dashboard = vm.Dashboard(pages=two_pages)
    return dashboard



"""Unit tests for hyphen.models.slider."""
import json

import plotly
import pytest
from dash import dcc, html
from pydantic import ValidationError

import vizro.models as vm


@pytest.fixture()
def expected_slider():
    return html.Div(
        [
            dcc.Store(
                "slider_id_callback_data",
                data={
                    "id": "slider_id",
                    "min": 0,
                    "max": 10,
                },
            ),
            html.P("Test title"),
            html.Div(
                [
                    dcc.Slider(
                        id="slider_id",
                        min=0,
                        max=10,
                        step=1,
                        marks={},
                        value=5,
                        included=False,
                        persistence=True,
                        className="slider_control",
                    ),
                    dcc.Input(
                        id="slider_id_text_value",
                        type="number",
                        placeholder="end",
                        min=0,
                        max=10,
                        value=5,
                        persistence=True,
                        className="slider_input_field_right",
                    ),
                    dcc.Store(id="slider_id_temp_store", storage_type="local"),
                ],
                className="slider_inner_container",
            ),
        ],
        className="selector_container",
        id="slider_id_outer",
    )


class TestSliderInstantiation:
    """Tests model instantiation."""

    def test_create_slider_mandatory(self):
        slider = vm.Slider()

        assert hasattr(slider, "id")
        assert slider.type == "slider"
        assert slider.min is None
        assert slider.max is None
        assert slider.step is None
        assert slider.marks is None
        assert slider.value is None
        assert slider.title is None
        assert slider.actions == []

    @pytest.mark.parametrize(
        "min, max",
        [
            (0, None),
            (None, 10),
            (0, 10),
        ],
    )
    def test_valid_min_max(self, min, max):
        slider = vm.Slider(min=min, max=max)

        assert slider.min == min
        assert slider.max == max

    def test_validate_max_invalid(self):
        with pytest.raises(
            ValidationError,
            match="Maximum value of slider is required to be larger than minimum value.",
        ):
            vm.Slider(min=10, max=0)

    @pytest.mark.parametrize(
        "value",
        [
            5,
            -5,
            0,
            6.5,
            -10,
            10,
        ],
    )
    def test_validate_slider_value_valid(self, value):
        slider = vm.Slider(min=-10, max=10, value=value)

        assert slider.value == value

    @pytest.mark.parametrize(
        "value",
        [
            11,
            -1,
        ],
    )
    def test_validate_slider_value_invalid(self, value):
        with pytest.raises(
            ValidationError,
            match="Please provide a valid value between the min and max value.",
        ):
            vm.Slider(min=0, max=10, value=value)

    @pytest.mark.parametrize(
        "step, expected",
        [(1, 1), (2.5, 2.5), (10, 10), (None, None), ("1", 1.0)],
    )
    def test_validate_step_valid(self, step, expected):
        slider = vm.Slider(min=0, max=10, step=step)

        assert slider.step == expected

    def test_validate_step_invalid(self):
        with pytest.raises(
            ValidationError,
            match="The step value of the slider must be less than or equal to the difference between max and min.",
        ):
            vm.Slider(min=0, max=10, step=11)

    def test_valid_marks_with_step(self):
        slider = vm.Slider(min=0, max=10, step=1)

        assert slider.marks == {}

    @pytest.mark.parametrize(
        "marks, step, expected",
        [
            ({2: "2", 4: "4", 6: "6"}, 1, {}),
            ({2: "2", 4: "4", 6: "6"}, None, {2: "2", 4: "4", 6: "6"}),
            ({}, 1, {}),
        ],
    )
    def test_step_precedence_over_marks(self, marks, step, expected):
        slider = vm.Slider(min=0, max=10, marks=marks, step=step)

        assert slider.marks == expected
        assert slider.step == step

    @pytest.mark.parametrize(
        "marks, expected",
        [
            ({i: str(i) for i in range(0, 10, 5)}, {i: str(i) for i in range(0, 10, 5)}),
            ({15: 15, 25: 25}, {15.0: "15", 25.0: "25"}),
            ({"15": 15, "25": 25}, {15.0: "15", 25.0: "25"}),
            (None, None),
        ],
    )
    def test_valid_marks(self, marks, expected):
        slider = vm.Slider(min=0, max=10, marks=marks)

        assert slider.marks == expected

    def test_invalid_marks(self):
        with pytest.raises(
            ValidationError,
            match="2 validation errors for Slider",
        ):
            vm.Slider(min=1, max=10, marks={"start": 0, "end": 10})

    @pytest.mark.parametrize("step, expected", [(1, {}), (None, None)])
    def test_set_default_marks(self, step, expected):
        slider = vm.Slider(min=0, max=10, step=step)
        assert slider.marks == expected

    @pytest.mark.parametrize(
        "title",
        [
            "test",
            1,
            1.0,
            """## Test header""",
            "",
        ],
    )
    def test_valid_title(self, title):
        slider = vm.Slider(title=title)

        assert slider.title == str(title)

    def test_set_action_via_validator(self, test_action_function):
        slider = vm.Slider(actions=[vm.Action(function=test_action_function)])
        actions_chain = slider.actions[0]

        assert actions_chain.trigger.component_property == "value"


class TestBuildMethod:
    def test_slider_build(self, expected_slider):
        slider = vm.Slider(min=0, max=10, step=1, value=5, id="slider_id", title="Test title")
        slider = slider.build()

        result = json.loads(json.dumps(slider, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_slider, cls=plotly.utils.PlotlyJSONEncoder))

        assert result == expected



"""Functionality to enable drop-in replacement that wraps plotly express figures.

Makes them compatible with the dashboard when you do `import vizro.plotly.express as px`.
Only plotly figures are wrapped; everything else is passed through unmodified, e.g. px.data.
"""
from typing import Any

import plotly.express as px

from vizro.models.types import capture


# TODO: is there a better way to see if the import is a chart? Don't want to check return type though. -> MS
# Might also want to define __dir__ or __all__ in order to facilitate IDE completion etc.
# TODO: type hints -> MS
def __getattr__(name: str) -> Any:
    px_name = getattr(px, name)
    try:
        return capture(mode="graph")(px_name) if px_name.__module__ == "plotly.express._chart_types" else px_name
    except AttributeError:
        return px_name



"""The model manager handles access to all `VizroBaseModel` instances used in a Vizro app."""
from __future__ import annotations

import random
import uuid
from typing import TYPE_CHECKING, Dict, Generator, NewType, Tuple, Type, TypeVar, cast

from vizro.managers._managers_utils import _state_modifier

if TYPE_CHECKING:
    from vizro.models import VizroBaseModel

rd = random.Random(0)

ModelID = NewType("ModelID", str)
Model = TypeVar("Model", bound="VizroBaseModel")


class DuplicateIDError(ValueError):
    """Useful for providing a more explicit error message when a model has id set automatically, e.g. Page."""

    pass


class ModelManager:
    def __init__(self):
        self.__models: Dict[ModelID, VizroBaseModel] = {}
        self._frozen_state = False

    @_state_modifier
    def __setitem__(self, model_id: ModelID, model: Model):
        if model_id in self.__models:
            raise DuplicateIDError(
                f"Model with id={model_id} already exists. Models must have a unique id across the whole dashboard. "
                f"If you are working from a Jupyter Notebook, please either restart the kernel, or "
                f"use 'from vizro.managers import model_manager; model_manager._reset()`."
            )
        self.__models[model_id] = model

    def __getitem__(self, model_id: ModelID) -> VizroBaseModel:
        # Do we need to return deepcopy(self.__models[model_id]) to avoid adjusting element by accident?
        return self.__models[model_id]

    def __iter__(self) -> Generator[ModelID, None, None]:
        """Iterates through all models.

        Note this yields model IDs rather key/value pairs to match the interface for a dictionary.
        """
        yield from self.__models

    def _items_with_type(self, model_type: Type[Model]) -> Generator[Tuple[ModelID, Model], None, None]:
        """Iterates through all models of type `model_type` (including subclasses)."""
        for model_id in self:
            if isinstance(self[model_id], model_type):
                yield model_id, cast(Model, self[model_id])

    @staticmethod
    def _generate_id() -> ModelID:
        return ModelID(str(uuid.UUID(int=rd.getrandbits(128))))

    def _reset(self):
        self.__init__()  # type: ignore[misc]


model_manager = ModelManager()



from copy import deepcopy

import plotly.graph_objects as go
import pytest

from vizro.models.types import capture


@pytest.fixture
def minimal_capture_chart(gapminder):
    @capture("graph")
    def _minimal_capture_chart(data_frame):
        return go.Figure()

    return _minimal_capture_chart(gapminder)


@pytest.mark.parametrize("chart_type", ["standard_px_chart", "minimal_capture_chart"])
def test_deepcopy_DashboardReadyFigure(chart_type, request):
    original = request.getfixturevalue(chart_type)
    copy = deepcopy(original)

    assert hasattr(copy, "_captured_callable")
    assert original._captured_callable is not copy._captured_callable
    assert original._captured_callable() == copy._captured_callable()



import random

import numpy as np
import pandas as pd
import pytest

import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro


@pytest.fixture
def dfs_with_shared_column():
    df1 = pd.DataFrame()
    df1["x"] = np.random.uniform(0, 10, 100)
    df1["y"] = np.random.uniform(0, 10, 100)
    df2 = df1.copy()

    df1["shared_column"] = np.random.uniform(0, 10, 100)
    df2["shared_column"] = random.choices(["CATEGORY 1", "CATEGORY 2"], k=100)

    return df1, df2


@pytest.fixture
def managers_one_page_two_graphs(gapminder):
    """Instantiates a simple model_manager and data_manager with a page, and two graph models and gapminder data."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(id="scatter_chart", figure=px.scatter(gapminder, x="lifeExp", y="gdpPercap")),
            vm.Graph(id="bar_chart", figure=px.bar(gapminder, x="country", y="gdpPercap")),
        ],
    )
    Vizro._pre_build()


@pytest.fixture
def managers_shared_column_different_dtype(dfs_with_shared_column):
    """Instantiates the managers with a page and two graphs sharing the same column but of different data types."""
    df1, df2 = dfs_with_shared_column
    vm.Page(
        id="graphs_with_shared_column",
        title="Page Title",
        components=[
            vm.Graph(figure=px.scatter(df1, x="x", y="y", color="shared_column")),
            vm.Graph(figure=px.scatter(df2, x="x", y="y", color="shared_column")),
        ],
    )
    Vizro._pre_build()



from pathlib import Path
from typing import Any, Dict, Optional, Union

from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from kedro.io import DataCatalog

from vizro.managers._data_manager import pd_LazyDataFrame


def catalog_from_project(
    project_path: Union[str, Path], env: Optional[str] = None, extra_params: Optional[Dict[str, Any]] = None
) -> DataCatalog:
    bootstrap_project(project_path)
    with KedroSession.create(
        project_path=project_path, env=env, save_on_close=False, extra_params=extra_params
    ) as session:
        return session.load_context().catalog


def datasets_from_catalog(catalog: DataCatalog) -> Dict[str, pd_LazyDataFrame]:
    datasets = {}
    for name in catalog.list():
        dataset = catalog._get_dataset(name, suggest=False)
        if "pandas" in dataset.__module__:
            datasets[name] = dataset.load
    return datasets



"""Unit tests for vizro.models.Graph."""
import json

import plotly
import plotly.graph_objects as go
import pytest
from dash import dcc
from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro.managers import data_manager
from vizro.models._action._action import Action
from vizro.models._components.graph import create_empty_fig


@pytest.fixture
def standard_go_chart(gapminder):
    return go.Figure(data=go.Scatter(x=gapminder["gdpPercap"], y=gapminder["lifeExp"], mode="markers"))


@pytest.fixture
def standard_px_chart_with_str_dataframe():
    return px.scatter(
        data_frame="gapminder",
        x="gdpPercap",
        y="lifeExp",
        size="pop",
        color="continent",
        hover_name="country",
        size_max=60,
    )


@pytest.fixture
def expected_empty_chart():
    figure = go.Figure()
    figure.add_trace(go.Scatter(x=[None], y=[None], showlegend=False, hoverinfo="none"))
    figure.update_layout(
        xaxis={"visible": False},
        yaxis={"visible": False},
        annotations=[{"text": "NO DATA", "showarrow": False, "font": {"size": 16}}],
    )
    return figure


@pytest.fixture
def expected_graph():
    return dcc.Loading(
        dcc.Graph(
            id="text_graph",
            figure=create_empty_fig(""),
            config={
                "autosizable": True,
                "frameMargins": 0,
                "responsive": True,
            },
            className="chart_container",
        ),
        color="grey",
        parent_className="chart_container",
    )


class TestDunderMethodsGraph:
    def test_create_graph_mandatory_only(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart)

        assert hasattr(graph, "id")
        assert graph.type == "graph"
        assert graph.figure == standard_px_chart._captured_callable
        assert graph.actions == []

    @pytest.mark.parametrize("id", ["id_1", "id_2"])
    def test_create_graph_mandatory_and_optional(self, standard_px_chart, id):
        graph = vm.Graph(
            figure=standard_px_chart,
            id=id,
            actions=[],
        )

        assert graph.id == id
        assert graph.type == "graph"
        assert graph.figure == standard_px_chart._captured_callable

    def test_mandatory_figure_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Graph()

    def test_failed_graph_with_wrong_figure(self, standard_go_chart):
        with pytest.raises(ValidationError, match="must provide a valid CapturedCallable object"):
            vm.Graph(
                figure=standard_go_chart,
            )

    def test_getitem_known_args(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart)

        assert graph["x"] == "gdpPercap"
        assert graph["type"] == "graph"

    def test_getitem_unknown_args(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart)
        with pytest.raises(KeyError):
            graph["unknown_args"]

    @pytest.mark.parametrize("title, expected", [(None, 24), ("Test", None)])
    def test_title_margin_adjustment(self, gapminder, title, expected):
        figure = vm.Graph(figure=px.bar(data_frame=gapminder, x="year", y="pop", title=title)).__call__()

        assert figure.layout.margin.t == expected
        assert figure.layout.template.layout.margin.t == 64
        assert figure.layout.template.layout.margin.l == 80
        assert figure.layout.template.layout.margin.b == 64
        assert figure.layout.template.layout.margin.r == 12

    def test_set_action_via_validator(self, standard_px_chart, test_action_function):
        graph = vm.Graph(figure=standard_px_chart, actions=[Action(function=test_action_function)])
        actions_chain = graph.actions[0]
        assert actions_chain.trigger.component_property == "clickData"


class TestProcessFigureDataFrame:
    def test_process_figure_data_frame_str_df(self, standard_px_chart_with_str_dataframe, gapminder):
        data_manager["gapminder"] = gapminder
        graph_with_str_df = vm.Graph(
            id="text_graph",
            figure=standard_px_chart_with_str_dataframe,
        )
        assert data_manager._get_component_data("text_graph").equals(gapminder)
        assert graph_with_str_df["data_frame"] == "gapminder"

    def test_process_figure_data_frame_df(self, standard_px_chart, gapminder):
        graph_with_df = vm.Graph(
            id="text_graph",
            figure=standard_px_chart,
        )
        assert data_manager._get_component_data("text_graph").equals(gapminder)
        with pytest.raises(KeyError, match="'data_frame'"):
            graph_with_df.figure["data_frame"]


class TestBuild:
    def test_create_empty_fig(self, expected_empty_chart):
        result = create_empty_fig("NO DATA")
        assert result == expected_empty_chart

    def test_graph_build(self, standard_px_chart, expected_graph):
        graph = vm.Graph(
            id="text_graph",
            figure=standard_px_chart,
        )

        result = json.loads(json.dumps(graph.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_graph, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
from vizro.actions import filter_interaction
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def callback_context_click_continent(request):
    """Mock dash.callback_context that represents a click on a continent data-point."""
    continent = request.param
    mock_callback_context = {
        "args_grouping": {
            "filters": [],
            "filter_interaction": [
                CallbackTriggerDict(
                    id="box_chart",
                    property="clickData",
                    value={
                        "points": [
                            {
                                "customdata": [continent],
                            }
                        ]
                    },
                    str_id="box_chart",
                    triggered=False,
                )
            ],
            "parameters": [],
            "theme_selector": CallbackTriggerDict(
                id="theme_selector",
                property="on",
                value=True,
                str_id="theme_selector",
                triggered=False,
            ),
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
class TestFilterInteraction:
    @pytest.mark.parametrize("callback_context_click_continent", ["Africa", "Europe"], indirect=True)
    def test_filter_interaction_without_targets_temporary_behavior(  # temporary fix, see below test
        self,
        callback_context_click_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["test_page"].components[0].actions = [vm.Action(id="test_action", function=filter_interaction())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result == {}

    @pytest.mark.xfail  # This is the desired behavior, ie when no target is provided, then all charts filtered
    @pytest.mark.parametrize(
        "callback_context_click_continent,target_scatter_filtered_continent,target_box_filtered_continent",
        [("Africa", "Africa", "Africa"), ("Europe", "Europe", "Europe"), ("Americas", "Americas", "Americas")],
        indirect=True,
    )
    def test_filter_interaction_without_targets_desired_behavior(
        self,
        callback_context_click_continent,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["test_page"].components[0].actions = [vm.Action(id="test_action", function=filter_interaction())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["scatter_chart"] == target_scatter_filtered_continent
        assert result["box_chart"] == target_box_filtered_continent

    @pytest.mark.parametrize(
        "callback_context_click_continent,target_scatter_filtered_continent",
        [("Africa", ["Africa"]), ("Europe", ["Europe"]), ("Americas", ["Americas"])],
        indirect=True,
    )
    def test_filter_interaction_with_one_target(
        self,
        callback_context_click_continent,
        target_scatter_filtered_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["test_page"].components[0].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart"]))
        ]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["scatter_chart"] == target_scatter_filtered_continent

    @pytest.mark.parametrize(
        "callback_context_click_continent,target_scatter_filtered_continent,target_box_filtered_continent",
        [
            ("Africa", ["Africa"], ["Africa"]),
            ("Europe", ["Europe"], ["Europe"]),
            ("Americas", ["Americas"], ["Americas"]),
        ],
        indirect=True,
    )
    def test_filter_interaction_with_two_target(
        self,
        callback_context_click_continent,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["test_page"].components[0].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["scatter_chart"] == target_scatter_filtered_continent
        assert result["box_chart"] == target_box_filtered_continent

    @pytest.mark.xfail  # This (or similar code) should raise a Value/Validation error explaining next steps
    @pytest.mark.parametrize("target", ["scatter_chart", ["scatter_chart"]])
    @pytest.mark.parametrize("callback_context_click_continent", ["Africa", "Europe"], indirect=True)
    def test_filter_interaction_with_invalid_targets(
        self,
        target,
        callback_context_click_continent,
    ):
        with pytest.raises(ValueError, match="Target invalid_target not found in model_manager."):
            # Add action to relevant component - here component[0] is the source_chart
            model_manager["test_page"].components[0].actions = [
                vm.Action(id="test_action", function=filter_interaction(targets=target))
            ]

    # TODO: Simplify parametrization, such that we have less repetitive code
    # TODO: Eliminate above xfails
    # TODO: Complement tests above with backend tests (currently the targets are also taken from model_manager!
    # see _get_action_callback_mapping eg.)



"""Types used in pydantic fields."""
# ruff: noqa: F821
from __future__ import annotations

import functools
import inspect
from copy import deepcopy
from typing import Any, Dict, List, Literal, Protocol, Union, runtime_checkable

from pydantic import Field, StrictBool
from pydantic.fields import ModelField
from pydantic.schema import SkipField
from typing_extensions import Annotated, TypedDict

from vizro.charts._charts_utils import _DashboardReadyFigure


# Used to describe _DashboardReadyFigure so we can keep CapturedCallable generic rather than referring to
# _DashboardReadyFigure explicitly.
@runtime_checkable
class _SupportsCapturedCallable(Protocol):
    _captured_callable: CapturedCallable


class CapturedCallable:
    """Stores a captured function call to use in a dashboard.

    Users do not need to instantiate this class directly. Instances are instead generated automatically
    through the [`capture`][vizro.models.types.capture] decorator. Some of the functionality is similar to
    `functools.partial`.

    Ready-to-use `CapturedCallable` instances are provided by Vizro. In this case refer to the user guide on
    [Charts/Graph][graph] or [Actions][pre-defined-actions] to see available choices.

    (Advanced) In case you would like to create your own `CapturedCallable`, please refer to the user guide on
    [custom charts](../user_guides/custom_charts.md) or [custom actions][custom-actions].
    """

    def __init__(self, function, /, *args, **kwargs):
        """Creates a new CapturedCallable object that will be able to re-run `function`.

        Partially binds *args and **kwargs to the function call.
        """
        self.__function = function
        self.__bound_arguments = inspect.signature(function).bind_partial(*args, **kwargs)

    def __call__(self, **kwargs):
        """Run the `function` with the initial arguments overridden by **kwargs.

        Note *args are not possible here, but you can still override positional arguments using argument name.
        """
        if not kwargs:
            return self.__function(*self.__bound_arguments.args, **self.__bound_arguments.kwargs)

        bound_arguments = deepcopy(self.__bound_arguments)
        bound_arguments.arguments.update(kwargs)
        # This looks like it should be self.__function(*bound_arguments.args, **bound_arguments.kwargs) as in the
        # Python documentation, but that leads to problems due to https://bugs.python.org/issue41745.
        return self.__function(**bound_arguments.arguments)

    def __getitem__(self, arg_name: str):
        """Gets the value of a bound argument."""
        return self.__bound_arguments.arguments[arg_name]

    def __delitem__(self, arg_name: str):
        """Deletes a bound argument."""
        del self.__bound_arguments.arguments[arg_name]

    @property
    def _arguments(self):
        # TODO: This is only used once in _get_parametrized_config and should be removed when that reference is
        #  removed.
        return self.__bound_arguments.arguments

    @classmethod
    def __get_validators__(cls):
        """Makes type compatible with pydantic model without needing arbitrary_types_allowed."""
        yield cls._parse_json

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any], field: ModelField):
        """Generates schema for field of this type."""
        raise SkipField(f"{cls.__name__} {field.name} is excluded from the schema.")

    @classmethod
    def _parse_json(
        cls,
        callable_config: Union[_SupportsCapturedCallable, CapturedCallable, Dict[str, Any]],
        field: ModelField,
    ) -> CapturedCallable:
        """Parses callable_config specification from JSON/YAML to a CapturedCallable.

        This uses the hydra syntax for _target_ but none of the other bits and we don't actually use hydra
        to implement it. In future, we might like to switch to using hydra's actual implementation
        which would allow nested functions (e.g. for transformers?) and to specify the path to a _target_ that lives
        outside of vizro.plotly_express. See https://hydra.cc/docs/advanced/instantiate_objects/overview/.
        """
        if isinstance(callable_config, CapturedCallable):
            # e.g. an action function that is already CapturedCallable
            return callable_config
        elif isinstance(callable_config, _SupportsCapturedCallable):
            # e.g. a _DashboardReadyFigure that has CapturedCallable in a property ._captured_callable
            return callable_config._captured_callable
        elif not isinstance(callable_config, dict):
            raise ValueError(
                "You must provide a valid CapturedCallable object. If you are using a plotly express figure, ensure "
                "that you are using `import vizro.plotly.express as px`. If you are using a custom figure or action, "
                "that your function uses the @capture decorator."
            )

        # Try to import function given in _target_ from the import_path property of the pydantic field.
        try:
            function_name = callable_config.pop("_target_")
        except KeyError as exc:
            raise ValueError(
                "CapturedCallable object must contain the key '_target_' that gives the target function."
            ) from exc

        import_path = field.field_info.extra["import_path"]
        try:
            function = getattr(import_path, function_name)
        except AttributeError as exc:
            raise ValueError(f"_target_={function_name} cannot be imported from {import_path.__name__}.") from exc

        # All the other items in figure are the keyword arguments to pass into function.
        function_kwargs = callable_config

        # It would seem natural to return cls(function, **function_kwargs) here, but the function is already decorated
        # with @capture, and so that would return a nested CapturedCallable.
        captured_callable = function(**function_kwargs)
        if isinstance(captured_callable, CapturedCallable):
            # e.g. an action function that is already CapturedCallable
            return captured_callable
        elif isinstance(captured_callable, _SupportsCapturedCallable):
            # e.g. a _DashboardReadyFigure that has CapturedCallable in a property ._captured_callable
            return captured_callable._captured_callable
        else:
            raise ValueError(f"_target_={function_name} must be wrapped in the @capture decorator.")

    # TODO-actions: Find a way how to compare CapturedCallable and function
    @property
    def _function(self):
        return self.__function


class capture:
    """Captures a function call to create a [`CapturedCallable`][vizro.models.types.CapturedCallable].

    This is used to add the functionality required to make graphs and actions work in a dashboard.
    Typically it should be used as a function decorator. There are two possible modes: `"graph"` and `"action"`.

    Examples:
        >>> @capture("graph")
        >>> def plot_function():
        >>>     ...
        >>> @capture("action")
        >>> def action_function():
        >>>     ...

    For further help on the use of `@capture("graph")`, you can refer to the guide on
    [custom charts](../user_guides/custom_charts.md).
    """

    def __init__(self, mode: Literal["graph", "action"]):
        """Instantiates the decorator to capture a function call. Valid modes are "graph" and "action"."""
        self._mode = mode

    def __call__(self, func, /):
        """Produces a CapturedCallable or _DashboardReadyFigure.

        mode="action" gives a CapturedCallable, while mode="graph" gives a _DashboardReadyFigure that contains a
        CapturedCallable. In both cases, the CapturedCallable is based on func and the provided *args and **kwargs.
        """
        if self._mode == "graph":
            # The more difficult case, where we need to still have a valid plotly figure that renders in a notebook.
            # Hence we attach the CapturedCallable as a property instead of returning it directly.
            # TODO: move point of checking that data_frame argument exists earlier on.
            @functools.wraps(func)
            def wrapped(*args, **kwargs) -> _DashboardReadyFigure:
                if "data_frame" not in inspect.signature(func).parameters:
                    raise ValueError(f"{func.__name__} must have data_frame argument to use capture('graph').")

                # We need to capture function upfront in order to find value of data_frame argument: since it could be
                # positional or keyword, this is much more robust than trying to get it out of arg or kwargs ourselves.
                captured_callable: CapturedCallable = CapturedCallable(func, *args, **kwargs)

                try:
                    captured_callable["data_frame"]
                except KeyError as exc:
                    raise ValueError(f"{func.__name__} must supply a value to data_frame argument.") from exc

                if isinstance(captured_callable["data_frame"], str):
                    # Enable running e.g. px.scatter("iris") from the Python API. Don't actually run the function
                    # because it won't get work as there's no data. It's vital we don't fetch data from the data manager
                    # yet either, because otherwise all lazy data will be loaded before the dashboard is started.
                    # This case is not relevant for the JSON/YAML API, which is handled separately through validation of
                    # CapturedCallable.
                    fig = _DashboardReadyFigure()
                else:
                    # Standard case for px.scatter(df: pd.DataFrame).
                    fig = func(*args, **kwargs)
                    fig.__class__ = _DashboardReadyFigure

                fig._captured_callable = captured_callable
                return fig

            return wrapped
        elif self._mode == "action":
            # The "normal" case where we just capture the function call.
            @functools.wraps(func)
            def wrapped(*args, **kwargs):
                # Note this is basically the same as partial(func, *args, **kwargs)
                return CapturedCallable(func, *args, **kwargs)

            return wrapped

        raise ValueError("Valid modes of the capture decorator are @capture('graph') and @capture('action').")


# Types used for selector values and options. Note the docstrings here are rendered on the API reference.
SingleValueType = Union[StrictBool, float, str]
"""Permissible value types for single-value selectors. Values are displayed as default."""
MultiValueType = Union[List[StrictBool], List[float], List[str]]
"""Permissible value types for multi-value selectors. Values are displayed as default."""


class OptionsDictType(TypedDict):
    """Permissible sub-type for OptionsType. Needs to be in the format of {"label": XXX, "value": XXX}."""

    label: str
    value: SingleValueType


OptionsType = Union[List[StrictBool], List[float], List[str], List[OptionsDictType]]
"""Permissible options types for selectors. Options are available choices for user to select from."""

# All the below types rely on models and so must use ForwardRef (i.e. "Checklist" rather than actual Checklist class).
SelectorType = Annotated[
    Union["Checklist", "Dropdown", "RadioItems", "RangeSlider", "Slider"],
    Field(
        discriminator="type",
        description="Selectors to be used inside a control.",
    ),
]
"""Discriminated union. Type of selector to be used inside a control: [`Checklist`][vizro.models.Checklist],
[`Dropdown`][vizro.models.Dropdown], [`RadioItems`][vizro.models.RadioItems],
[`RangeSlider`][vizro.models.RangeSlider] or [`Slider`][vizro.models.Slider]."""

_FormComponentType = Annotated[
    Union[SelectorType, "Button", "UserInput"],
    Field(
        discriminator="type",
        description="Components that can be used to receive user input within a form'.",
    ),
]

ControlType = Annotated[
    Union["Filter", "Parameter"],
    Field(
        discriminator="type",
        description="Control that affects components on the page.",
    ),
]
"""Discriminated union. Type of control that affects components on the page: [`Filter`][vizro.models.Filter] or
[`Parameter`][vizro.models.Parameter]."""

ComponentType = Annotated[
    Union["Button", "Card", "Graph"],
    Field(
        discriminator="type",
        description="Component that makes up part of the layout on the page.",
    ),
]
"""Discriminated union. Type of component that makes up part of the layout on the page:
[`Button`][vizro.models.Button], [`Card`][vizro.models.Card] or [`Graph`][vizro.models.Graph]."""

# Types used for pages values in the Navigation model.
NavigationPagesType = Annotated[
    Union[List[str], Dict[str, List[str]]],
    Field(
        None, description="List of Page IDs or dict mapping of Page IDs and titles (for hierarchical sub-navigation)"
    ),
]
"""Permissible value types for page attribute. Values are displayed as default."""



"""Example to show dashboard configuration specified as a YAML file."""
import os
from pathlib import Path

import pandas as pd
import yaml

import vizro.plotly.express as px
from vizro import Vizro
from vizro.managers import data_manager
from vizro.models import Dashboard


def retrieve_gapminder():
    """This is a function that returns gapminder data."""
    return px.data.gapminder()


def retrieve_gapminder_year(year: int):
    """This is a function that returns gapminder data for a year."""
    return px.data.gapminder().query(f"year == {year}")


def retrieve_gapminder_continent_comparison():
    """This is a function adds aggregated continent information to gapminder data."""
    df_gapminder = px.data.gapminder()
    df_gapminder_agg = px.data.gapminder()

    df_gapminder_agg["lifeExp"] = df_gapminder_agg.groupby(by=["continent", "year"])["lifeExp"].transform("mean")
    df_gapminder_agg["gdpPercap"] = df_gapminder_agg.groupby(by=["continent", "year"])["gdpPercap"].transform("mean")
    df_gapminder_agg["pop"] = df_gapminder_agg.groupby(by=["continent", "year"])["pop"].transform("sum")

    df_gapminder["data"] = "Country"
    df_gapminder_agg["data"] = "Continent"

    df_gapminder_comp = pd.concat([df_gapminder_agg, df_gapminder], ignore_index=True)

    return df_gapminder_comp


def retrieve_avg_gapminder():
    """This is a function that returns aggregated gapminder data."""
    df = px.data.gapminder()
    mean = (
        df.groupby(by=["continent", "year"]).agg({"lifeExp": "mean", "pop": "mean", "gdpPercap": "mean"}).reset_index()
    )
    return mean


def retrieve_avg_gapminder_year(year: int):
    """This is a function that returns aggregated gapminder data for a specific year."""
    return retrieve_avg_gapminder().query(f"year == {year}")


# If you're not interested in lazy loading then you could just do data_manager["gapminder"] = px.data.gapminder()
data_manager["gapminder"] = retrieve_gapminder
data_manager["gapminder_2007"] = lambda: retrieve_gapminder_year(2007)
data_manager["gapminder_avg"] = retrieve_avg_gapminder
data_manager["gapminder_avg_2007"] = lambda: retrieve_avg_gapminder_year(2007)
data_manager["gapminder_country_analysis"] = retrieve_gapminder_continent_comparison

dashboard = yaml.safe_load(Path("dashboard.yaml").read_text(encoding="utf-8"))
dashboard = Dashboard(**dashboard)

if __name__ == "__main__":
    Vizro._user_assets_folder = os.path.abspath("../assets")
    Vizro().build(dashboard).run()



import numpy as np
import pytest
from pydantic import ValidationError

import vizro.models as vm
from vizro.models._layout import GAP_DEFAULT, MIN_DEFAULT, ColRowGridLines
from vizro.models._models_utils import get_unique_grid_component_ids


class TestLayoutInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_layout_mandatory_only(self):
        layout = vm.Layout(grid=[[0, 1], [0, 2]])
        assert hasattr(layout, "id")
        assert layout.grid == [[0, 1], [0, 2]]
        assert layout.col_gap == GAP_DEFAULT
        assert layout.row_gap == GAP_DEFAULT
        assert layout.col_min_width == MIN_DEFAULT
        assert layout.row_min_height == MIN_DEFAULT
        assert layout.component_grid_lines == [
            ColRowGridLines(col_start=1, col_end=2, row_start=1, row_end=3),
            ColRowGridLines(col_start=2, col_end=3, row_start=1, row_end=2),
            ColRowGridLines(col_start=2, col_end=3, row_start=2, row_end=3),
        ]

    @pytest.mark.parametrize("test_gap", ["0px", "4px", "8px"])
    def test_create_layout_mandatory_and_optional(self, test_gap):
        layout = vm.Layout(
            grid=[[0, 1], [0, 2]], col_gap=test_gap, row_gap=test_gap, col_min_width=test_gap, row_min_height=test_gap
        )

        assert hasattr(layout, "id")
        assert layout.grid == [[0, 1], [0, 2]]
        assert layout.col_gap == test_gap
        assert layout.row_gap == test_gap
        assert layout.col_min_width == test_gap
        assert layout.row_min_height == test_gap
        assert layout.component_grid_lines == [
            ColRowGridLines(col_start=1, col_end=2, row_start=1, row_end=3),
            ColRowGridLines(col_start=2, col_end=3, row_start=1, row_end=2),
            ColRowGridLines(col_start=2, col_end=3, row_start=2, row_end=3),
        ]

    def test_mandatory_grid_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Layout()


class TestMalformedGrid:
    """Tests validators to detect malformed grid."""

    @pytest.mark.parametrize(
        "grid",
        [
            0,
            [0],
            [-1],
            [[0], 1],
            [0, 1, 0],
            [0, 1, 3],
        ],
    )
    def test_invalid_input_type(self, grid):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            vm.Layout(grid=grid)

    @pytest.mark.parametrize(
        "grid",
        [
            [["Table Graph"], ["Table ."]],
            [["A"], ["B"], ["C"]],
        ],
    )
    def test_invalid_input_value(self, grid):
        with pytest.raises(ValidationError, match="value is not a valid integer"):
            vm.Layout(grid=grid)

    @pytest.mark.parametrize(
        "grid",
        [
            [[0], [1], [0]],
            [[0, 1], [2, 0]],
            [[2, 0], [0, 1]],
            [[0, 1], [0, 0]],
            [[0, 0], [1, 0]],
            [[-1, 0], [0, -1]],
            [[-1, 0, 1], [0, -1, 1]],
            [[-1, 0, 1], [1, -1, 1]],
        ],
    )
    def test_invalid_grid_area(self, grid):
        with pytest.raises(ValidationError, match="Grid areas must be rectangular and not overlap!"):
            vm.Layout(grid=grid)

    @pytest.mark.parametrize(
        "grid",
        [
            [[1, 2], [3, 4]],
            [[0, 2], [3, 4]],
            [[6, 4], [2, 5]],
        ],
    )
    def test_invalid_int_sequence(self, grid):
        with pytest.raises(ValidationError, match="Grid must contain consecutive integers starting from 0."):
            vm.Layout(grid=grid)

    @pytest.mark.parametrize(
        "grid",
        [
            [[0, 1, 2], [3, 4]],
            [[0, 1], [2, 3, 4]],
        ],
    )
    def test_invalid_list_length(self, grid):
        with pytest.raises(ValidationError, match="All rows must be of same length."):
            vm.Layout(grid=grid)


class TestWorkingGrid:
    """Checks if working grid pass all validations."""

    @pytest.mark.parametrize(
        "grid",
        [
            [[0]],
            [[0], [1]],
            [[0, 1, 2]],
            [[0.0, 1.0, 2.0]],
            [[0, 1], [0, 2]],
            [[1, 0], [1, 0]],
            [[0, 0], [0, 0]],
            [[0, 1, 2], [3, 3, 3]],
            [[i] for i in range(10)],
            [[-1, 0], [1, 0]],
            [[-1, 0], [1, -1]],
            [[-1, -1, -1, -1], [0, 1, 2, 3]],
        ],
    )
    def test_working_grid(self, grid):
        try:
            vm.Layout(grid=grid)
        except ValidationError as ve:
            assert False, f"{grid} raised a value error {ve}."


@pytest.mark.parametrize("grid", [[[0, -1], [1, 2]], [[0, -1, 1, 2]], [[-1, -1, -1], [0, 1, 2]]])
def test_get_unique_grid_component_ids(grid):
    result = get_unique_grid_component_ids(grid)
    expected = np.array([0, 1, 2])

    assert isinstance(result, np.ndarray)
    assert (result == expected).all()



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
import vizro.plotly.express as px
from vizro._constants import PARAMETER_ACTION_PREFIX
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def target_scatter_parameter_y(request, gapminder_2007, scatter_params):
    y = request.param
    scatter_params["y"] = y
    return px.scatter(gapminder_2007, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_parameter_y(request, gapminder_2007, box_params):
    y = request.param
    box_params["y"] = y
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24)


@pytest.fixture
def target_scatter_parameter_y_and_x(request, gapminder_2007, scatter_params):
    y, x = request.param
    scatter_params["y"] = y
    scatter_params["x"] = x
    return px.scatter(gapminder_2007, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_parameter_y_and_x(request, gapminder_2007, box_params):
    y, x = request.param
    box_params["y"] = y
    box_params["x"] = x
    return px.box(gapminder_2007, **box_params).update_layout(margin_t=24)


@pytest.fixture
def callback_context_parameter_y(request):
    """Mock dash.callback_context that represents y-axis Parameter value selection."""
    y = request.param
    mock_callback_context = {
        "args_grouping": {
            "filter_interaction": [],
            "filters": [],
            "parameters": [
                CallbackTriggerDict(
                    id="y_parameter",
                    property="value",
                    value=y,
                    str_id="y_parameter",
                    triggered=False,
                )
            ],
            "theme_selector": CallbackTriggerDict(
                id="theme_selector",
                property="on",
                value=True,
                str_id="theme_selector",
                triggered=False,
            ),
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.fixture
def callback_context_parameter_y_and_x(request):
    """Mock dash.callback_context that represents y-axis Parameter value selection."""
    y, x = request.param
    mock_callback_context = {
        "args_grouping": {
            "filter_interaction": [],
            "filters": [],
            "parameters": [
                CallbackTriggerDict(
                    id="y_parameter",
                    property="value",
                    value=y,
                    str_id="y_parameter",
                    triggered=False,
                ),
                CallbackTriggerDict(
                    id="x_parameter",
                    property="value",
                    value=x,
                    str_id="x_parameter",
                    triggered=False,
                ),
            ],
            "theme_selector": CallbackTriggerDict(
                id="theme_selector",
                property="on",
                value=True,
                str_id="theme_selector",
                triggered=False,
            ),
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
class TestParameter:
    @pytest.mark.parametrize(
        "callback_context_parameter_y, target_scatter_parameter_y",
        [("pop", "pop"), ("gdpPercap", "gdpPercap")],
        indirect=True,
    )
    def test_one_parameter_one_target(
        self,
        callback_context_parameter_y,
        target_scatter_parameter_y,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()

        assert result["scatter_chart"] == target_scatter_parameter_y
        assert "box_chart" not in result

    @pytest.mark.parametrize(
        "callback_context_parameter_y, target_scatter_parameter_y, target_box_parameter_y",
        [("pop", "pop", "pop"), ("gdpPercap", "gdpPercap", "gdpPercap")],
        indirect=True,
    )
    def test_one_parameter_multiple_targets(
        self,
        callback_context_parameter_y,
        target_scatter_parameter_y,
        target_box_parameter_y,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter",
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        model_manager["test_page"].controls = [y_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter"].function()

        assert result["scatter_chart"] == target_scatter_parameter_y
        assert result["box_chart"] == target_box_parameter_y

    @pytest.mark.parametrize(
        "callback_context_parameter_y_and_x, target_scatter_parameter_y_and_x",
        [(["pop", "continent"], ["pop", "continent"]), (["gdpPercap", "country"], ["gdpPercap", "country"])],
        indirect=True,
    )
    def test_multiple_parameters_one_target(
        self,
        callback_context_parameter_y_and_x,
        target_scatter_parameter_y_and_x,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_x"].function()

        assert result["scatter_chart"] == target_scatter_parameter_y_and_x
        assert "box_chart" not in result

    @pytest.mark.parametrize(
        "callback_context_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x",
        [
            (["pop", "continent"], ["pop", "continent"], ["pop", "continent"]),
            (["gdpPercap", "country"], ["gdpPercap", "country"], ["gdpPercap", "country"]),
        ],
        indirect=True,
    )
    def test_multiple_parameters_multiple_targets(
        self,
        callback_context_parameter_y_and_x,
        target_scatter_parameter_y_and_x,
        target_box_parameter_y_and_x,
    ):
        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.x", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_x"].function()

        assert result["scatter_chart"] == target_scatter_parameter_y_and_x
        assert result["box_chart"] == target_box_parameter_y_and_x

    @pytest.mark.parametrize(
        "callback_context_parameter_y_and_x, target_scatter_parameter_y_and_x, target_box_parameter_y_and_x",
        [
            (["pop", "continent"], ["pop", "pop"], ["continent", "continent"]),
            (["gdpPercap", "country"], ["gdpPercap", "gdpPercap"], ["country", "country"]),
        ],
        indirect=True,
    )
    def test_one_parameter_per_target_multiple_attributes(
        self,
        callback_context_parameter_y_and_x,
        target_scatter_parameter_y_and_x,
        target_box_parameter_y_and_x,
    ):
        # Creating and adding a Parameter object to the existing Page
        scatter_parameter = vm.Parameter(
            id="test_parameter_scatter",
            targets=["scatter_chart.y", "scatter_chart.x"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        box_parameter = vm.Parameter(
            id="test_parameter_box",
            targets=["box_chart.y", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [scatter_parameter, box_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        scatter_parameter.pre_build()
        box_parameter.pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_scatter"].function()
        assert result["scatter_chart"] == target_scatter_parameter_y_and_x
        assert "box_chart" not in target_scatter_parameter_y_and_x

        result = model_manager[f"{PARAMETER_ACTION_PREFIX}_test_parameter_box"].function()
        assert result["box_chart"] == target_box_parameter_y_and_x
        assert "scatter_chart" not in target_scatter_parameter_y_and_x



"""Pre-defined action function "filter_interaction" to be re-used in `action` parameter of VizroBaseModels."""

from typing import Any, Dict, List, Optional

from dash import ctx

from vizro.actions._actions_utils import (
    _get_modified_page_charts,
)
from vizro.models.types import capture


@capture("action")
def filter_interaction(
    targets: Optional[List[str]] = None,
    **inputs: Dict[str, Any],
) -> Dict[str, Any]:
    """Filters targeted charts/components on page by clicking on data points of the source chart.

    To set up filtering on specific columns of the target chart(s), include these columns in the 'custom_data'
    parameter of the source chart e.g. `px.bar(..., custom_data=["species", "sepal_length"])`

    Args:
        targets: List of target component ids to filter by chart interaction. If missing, will target all valid
            components on page. Defaults to None.
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Returns:
        Dict mapping target component ids to modified charts/components e.g. {'my_scatter': Figure({})}
    """
    return _get_modified_page_charts(
        targets=targets,
        ctds_filter=ctx.args_grouping["filters"],
        ctds_filter_interaction=ctx.args_grouping["filter_interaction"],
        ctds_parameters=ctx.args_grouping["parameters"],
        ctd_theme=ctx.args_grouping["theme_selector"],
    )



"""Unit tests for vizro.models.Accordion."""
import json

import plotly
import pytest
from pydantic import ValidationError

import vizro.models as vm
from vizro.models._navigation._accordion import Accordion


@pytest.mark.usefixtures("dashboard_build")
class TestAccordionInstantiation:
    """Tests accordion model instantiation."""

    def test_create_accordion_default(self):
        accordion = Accordion(id="accordion_id")
        assert accordion.id == "accordion_id"
        assert accordion.pages is None

    def test_create_accordion_pages_as_list(self, pages_as_list):
        accordion = Accordion(pages=pages_as_list, id="accordion_id")
        assert accordion.id == "accordion_id"
        assert accordion.pages == pages_as_list

    def test_create_accordion_pages_as_dict(self, pages_as_dict):
        accordion = Accordion(pages=pages_as_dict, id="accordion_id")
        assert accordion.id == "accordion_id"
        assert accordion.pages == pages_as_dict

    def test_field_invalid_pages_input_type(self):
        with pytest.raises(ValidationError, match="2 validation errors for Accordion"):
            Accordion(pages=[vm.Button()])

    def test_field_invalid_pages_empty_list(self):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            Accordion(pages=[])


@pytest.mark.usefixtures("dashboard_build")
class TestAccordionBuild:
    """Tests accordion build method."""

    @pytest.mark.parametrize("pages", [["Page 1", "Page 2"], None])
    def test_accordion_build_default(self, pages, accordion_from_page_as_list):
        accordion = Accordion(pages=pages, id="accordion_list").build()
        result = json.loads(json.dumps(accordion, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(accordion_from_page_as_list, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    def test_accordion_build_pages_as_list(self, pages_as_list, accordion_from_page_as_list):
        accordion = Accordion(pages=pages_as_list, id="accordion_list").build()
        result = json.loads(json.dumps(accordion, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(accordion_from_page_as_list, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    def test_accordion_build_pages_as_dict(self, pages_as_dict, accordion_from_pages_as_dict):
        accordion = Accordion(pages=pages_as_dict, id="accordion_dict").build()
        result = json.loads(json.dumps(accordion, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(accordion_from_pages_as_dict, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    def test_accordion_build_single_page_accordion(self):
        accordion = Accordion(pages=["Page 1"], id="single_accordion").build()
        assert accordion is None

    def test_navigation_not_all_pages_included(self, dashboard_build):
        with pytest.warns(UserWarning):
            Accordion(pages=["Page 1"])



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
from vizro import Vizro
from vizro.actions import export_data, filter_interaction
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def target_scatter_filtered_pop_filter_interaction_continent(request, gapminder_2007):
    pop_filter, continent_filter_interaction = request.param
    data = gapminder_2007
    if pop_filter:
        data = data[data["pop"].between(pop_filter[0], pop_filter[1], inclusive="both")]
    if continent_filter_interaction:
        data = data[data["continent"].isin(continent_filter_interaction)]
    return data


@pytest.fixture
def target_box_filtered_pop(request, gapminder_2007):
    pop_filter = request.param
    data = gapminder_2007
    if pop_filter:
        data = data[data["pop"].between(pop_filter[0], pop_filter[1], inclusive="both")]
    return data


@pytest.fixture
def managers_one_page_without_graphs_one_button():
    """Instantiates a simple model_manager and data_manager with a page, and no graphs."""
    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[vm.Button(id="button")],
    )
    Vizro._pre_build()


@pytest.fixture
def callback_context_export_data(request):
    """Mock dash.callback_context that represents on page load."""
    targets, pop_filter, continent_filter_interaction = request.param
    mock_callback_context = {
        "args_grouping": {
            "filters": [
                CallbackTriggerDict(
                    id="pop_filter",
                    property="value",
                    value=pop_filter,
                    str_id="pop_filter",
                    triggered=False,
                )
            ]
            if pop_filter
            else [],
            "filter_interaction": [
                CallbackTriggerDict(
                    id="box_chart",
                    property="clickData",
                    value={
                        "points": [
                            {
                                "customdata": [continent_filter_interaction],
                            }
                        ]
                    },
                    str_id="box_chart",
                    triggered=False,
                )
            ]
            if continent_filter_interaction
            else [],
        },
        "outputs_list": [
            {"id": {"action_id": "test_action", "target_id": target, "type": "download-dataframe"}, "property": "data"}
            for target in targets
        ],
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


class TestExportData:
    @pytest.mark.usefixtures("managers_one_page_without_graphs_one_button")
    @pytest.mark.parametrize("callback_context_export_data", [([[], None, None])], indirect=True)
    def test_no_graphs_no_targets(self, callback_context_export_data):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result == {}

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data", [([["scatter_chart", "box_chart"], None, None])], indirect=True
    )
    def test_graphs_no_targets(self, callback_context_export_data, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["download-dataframe_scatter_chart"]["filename"] == "scatter_chart.csv"
        assert result["download-dataframe_scatter_chart"]["content"] == gapminder_2007.to_csv(index=False)

        assert result["download-dataframe_box_chart"]["filename"] == "box_chart.csv"
        assert result["download-dataframe_box_chart"]["content"] == gapminder_2007.to_csv(index=False)

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize("callback_context_export_data", [(["scatter_chart"], None, None)], indirect=True)
    def test_one_target(self, callback_context_export_data, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [vm.Action(id="test_action", function=export_data(targets=["scatter_chart"]))]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["download-dataframe_scatter_chart"]["filename"] == "scatter_chart.csv"
        assert result["download-dataframe_scatter_chart"]["content"] == gapminder_2007.to_csv(index=False)

        assert "download-dataframe_box_chart" not in result

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data", [(["scatter_chart", "box_chart"], None, None)], indirect=True
    )
    def test_multiple_targets(self, callback_context_export_data, gapminder_2007):
        # Add action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["download-dataframe_scatter_chart"]["filename"] == "scatter_chart.csv"
        assert result["download-dataframe_scatter_chart"]["content"] == gapminder_2007.to_csv(index=False)

        assert result["download-dataframe_box_chart"]["filename"] == "box_chart.csv"
        assert result["download-dataframe_box_chart"]["content"] == gapminder_2007.to_csv(index=False)

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize("callback_context_export_data", [(["invalid_target_id"], None, None)], indirect=True)
    def test_invalid_target(
        self,
        callback_context_export_data,
    ):
        # Add action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["invalid_target_id"]))
        ]

        with pytest.raises(ValueError, match="Component 'invalid_target_id' does not exist."):
            # Run action by picking the above added action function and executing it with ()
            model_manager["test_action"].function()

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data, "
        "target_scatter_filtered_pop_filter_interaction_continent, "
        "target_box_filtered_pop",
        [
            (
                [["scatter_chart", "box_chart"], [10**6, 10**7], None],
                [[10**6, 10**7], None],
                [10**6, 10**7],
            ),
            ([["scatter_chart", "box_chart"], None, "Africa"], [None, ["Africa"]], None),
            (
                [["scatter_chart", "box_chart"], [10**6, 10**7], "Africa"],
                [[10**6, 10**7], ["Africa"]],
                [10**6, 10**7],
            ),
        ],
        indirect=True,
    )
    def test_multiple_targets_with_filter_and_filter_interaction(
        self,
        callback_context_export_data,
        target_scatter_filtered_pop_filter_interaction_continent,
        target_box_filtered_pop,
    ):
        # Creating and adding a Filter object to the existing Page
        pop_filter = vm.Filter(column="pop", selector=vm.RangeSlider(id="pop_filter"))
        model_manager["test_page"].controls = [pop_filter]
        # Adds a default _filter Action to the filter selector objects
        pop_filter.pre_build()

        # Add filter_interaction Action to scatter_chart component
        model_manager["box_chart"].actions = [
            vm.Action(id="filter_interaction", function=filter_interaction(targets=["scatter_chart"]))
        ]

        # Add export_data action to relevant component
        model_manager["button"].actions = [
            vm.Action(id="test_action", function=export_data(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added export_data action function and executing it with ()
        result = model_manager["test_action"].function()

        assert result["download-dataframe_scatter_chart"]["filename"] == "scatter_chart.csv"
        assert result["download-dataframe_scatter_chart"][
            "content"
        ] == target_scatter_filtered_pop_filter_interaction_continent.to_csv(index=False)

        assert result["download-dataframe_box_chart"]["filename"] == "box_chart.csv"
        assert result["download-dataframe_box_chart"]["content"] == target_box_filtered_pop.to_csv(index=False)

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_export_data", [(["scatter_chart", "box_chart"], None, None)], indirect=True
    )
    def test_invalid_file_format(
        self,
        callback_context_export_data,
    ):
        # Add action to relevant component
        model_manager["button"].actions = [
            vm.Action(
                id="test_action",
                function=export_data(
                    targets=["scatter_chart", "box_chart"],
                    file_format="invalid_file_format",
                ),
            )
        ]

        with pytest.raises(
            ValueError, match='Unknown "file_format": invalid_file_format.' ' Known file formats: "csv", "xlsx".'
        ):
            # Run action by picking the above added action function and executing it with ()
            model_manager["test_action"].function()



"""Contains utilities to create required dash callbacks for the action loop."""
import logging
from typing import List

from dash import ClientsideFunction, Input, Output, State, clientside_callback

from vizro.actions._action_loop._action_loop_utils import (
    _get_actions_chains_on_registered_pages,
    _get_actions_on_registered_pages,
)

logger = logging.getLogger(__name__)


def _build_action_loop_callbacks() -> None:
    """Creates all required dash callbacks for the action loop."""
    actions_chains = _get_actions_chains_on_registered_pages()
    actions = _get_actions_on_registered_pages()

    if not actions_chains:
        return

    gateway_inputs: List[Input] = []
    for actions_chain in actions_chains:
        # Callback that enables gateway callback to work in the multiple page app
        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="trigger_to_global_store"),
            Output({"type": "gateway_input", "trigger_id": actions_chain.id}, "data"),
            Input(
                component_id=actions_chain.trigger.component_id,
                component_property=actions_chain.trigger.component_property,
            ),
            State({"type": "gateway_input", "trigger_id": actions_chain.id}, "data"),
            prevent_initial_call=True,
        )

        gateway_inputs.append(
            Input(
                component_id={"type": "gateway_input", "trigger_id": actions_chain.id},
                component_property="data",
            )
        )

    # Determines the final sequence of actions to be triggered.
    clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="gateway"),
        output=[Output("remaining_actions", "data")]
        + [Output({"type": "action_trigger", "action_name": action.id}, "data") for action in actions],
        inputs=[
            State("remaining_actions", "data"),
            State("trigger_to_actions_chain_mapper", "data"),
            State("action_trigger_actions_id", "data"),
            Input("cycle_breaker_div", "n_clicks"),
            *gateway_inputs,
        ],
        prevent_initial_call=True,
    )

    # Callback that triggers the next iteration
    clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="after_action_cycle_breaker"),
        Output("cycle_breaker_empty_output_store", "data"),
        Input("action_finished", "data"),
        prevent_initial_call=True,
    )



import json
from collections import OrderedDict

import dash
import dash_bootstrap_components as dbc
import plotly
import pytest
from dash import html
from pydantic import ValidationError

import vizro
import vizro.models as vm
from vizro import Vizro
from vizro.actions._action_loop._action_loop import ActionLoop
from vizro.managers import model_manager
from vizro.models._dashboard import create_layout_page_404, update_theme


@pytest.fixture()
def dashboard_container():
    return dbc.Container(
        id="dashboard_container",
        children=[
            html.Div(id=f"vizro_version_{vizro.__version__}"),
            *ActionLoop._create_app_callbacks(),
            dash.page_container,
        ],
        className="vizro_dark",
        fluid=True,
    )


@pytest.fixture()
def mock_page_registry():
    return OrderedDict(
        {
            "Page 1": {
                "module": "Page 1",
                "supplied_path": "/",
                "path_template": None,
                "path": "/",
                "supplied_name": "Page 1",
                "name": "Page 1",
                "supplied_title": None,
                "title": "Page 1",
                "description": "",
                "order": 0,
                "supplied_order": 0,
                "supplied_layout": model_manager["Page 1"].build,
                "supplied_image": None,
                "image": None,
                "image_url": None,
                "redirect_from": None,
                "layout": model_manager["Page 1"].build,
                "relative_path": "/",
            },
            "Page 2": {
                "module": "Page 2",
                "supplied_path": "/page-2",
                "path_template": None,
                "path": "/page-2",
                "supplied_name": "Page 2",
                "name": "Page 2",
                "supplied_title": None,
                "title": "Page 2",
                "description": "",
                "order": 1,
                "supplied_order": 1,
                "supplied_layout": model_manager["Page 2"].build,
                "supplied_image": None,
                "image": None,
                "image_url": None,
                "redirect_from": None,
                "layout": model_manager["Page 2"].build,
                "relative_path": "/page-2",
            },
            "not_found_404": {
                "module": "not_found_404",
                "supplied_path": None,
                "path_template": None,
                "path": "/not-found-404",
                "supplied_name": None,
                "name": "Not found 404",
                "supplied_title": None,
                "title": "Not found 404",
                "description": "",
                "order": None,
                "supplied_order": None,
                "supplied_layout": create_layout_page_404(),
                "supplied_image": None,
                "image": None,
                "image_url": None,
                "redirect_from": None,
                "layout": create_layout_page_404(),
                "relative_path": "/not-found-404",
            },
        }
    )


class TestDashboardInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_dashboard_mandatory_only(self, two_pages):
        dashboard = vm.Dashboard(pages=two_pages)
        assert hasattr(dashboard, "id")
        assert dashboard.pages == two_pages
        assert dashboard.theme == "vizro_dark"
        assert dashboard.title is None

    def test_create_dashboard_mandatory_and_optional(self, two_pages):
        dashboard = vm.Dashboard(pages=two_pages, theme="vizro_light", title="Vizro")
        assert hasattr(dashboard, "id")
        assert dashboard.pages == two_pages
        assert dashboard.theme == "vizro_light"
        assert dashboard.title == "Vizro"

    def test_mandatory_pages_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Dashboard()

    def test_field_invalid_pages_empty_list(self):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.Dashboard(pages=[])

    def test_field_invalid_pages_input_type(self):
        with pytest.raises(ValidationError, match="4 validation errors for Dashboard"):
            vm.Dashboard(pages=[vm.Button()])

    def test_field_invalid_theme_input_type(self, two_pages):
        with pytest.raises(ValidationError, match="unexpected value; permitted: 'vizro_dark', 'vizro_light'"):
            vm.Dashboard(pages=two_pages, theme="not_existing")


class TestDashboardBuild:
    """Tests dashboard build method."""

    def test_dashboard_container(self, dashboard, dashboard_container):
        app = Vizro().build(dashboard)
        result = json.loads(json.dumps(app.dash.layout, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(dashboard_container, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    def test_dashboard_page_registry(self, dashboard, mock_page_registry):
        Vizro().build(dashboard)
        result = dash.page_registry
        expected = mock_page_registry
        # Str conversion required as comparison of OrderedDict values result in False otherwise
        assert str(result.items()) == str(expected.items())


def test_create_layout_page_404():
    result = create_layout_page_404()
    result_image = result.children[0]
    result_div = result.children[1]

    assert isinstance(result, html.Div)
    assert isinstance(result_image, html.Img)
    assert isinstance(result_div, html.Div)


@pytest.mark.parametrize("on, expected", [(True, "vizro_dark"), (False, "vizro_light")])
def test_update_theme(on, expected):
    result = update_theme(on)
    assert result == expected



"""Unit tests for vizro.managers.data_manager."""
import pandas as pd
import pytest

from vizro.managers._data_manager import DataManager


class TestDataManager:
    def setup_method(self):
        self.data_manager = DataManager()
        self.data = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]})

    def test_add_dataframe(self):
        dataset_name = "test_dataset"
        component_id = "component_id_a"
        self.data_manager[dataset_name] = self.data
        self.data_manager._add_component(component_id, dataset_name)
        assert self.data_manager._get_component_data(component_id).equals(self.data)

    def test_add_lazy_dataframe(self):
        dataset_name = "test_lazy_dataset"
        component_id = "component_id_b"

        def lazy_data():
            return self.data

        self.data_manager[dataset_name] = lazy_data
        self.data_manager._add_component(component_id, dataset_name)
        assert self.data_manager._get_component_data(component_id).equals(lazy_data())

    def test_add_existing_dataset(self):
        dataset_name = "existing_dataset"
        self.data_manager[dataset_name] = self.data
        with pytest.raises(ValueError):
            self.data_manager[dataset_name] = self.data

    def test_add_invalid_dataset(self):
        dataset_name = "invalid_dataset"
        invalid_data = "not_a_dataframe"
        with pytest.raises(TypeError):
            self.data_manager[dataset_name] = invalid_data

    def test_add_component_to_nonexistent_dataset(self):
        component_id = "test_component"
        dataset_name = "nonexistent_dataset"
        with pytest.raises(KeyError):
            self.data_manager._add_component(component_id, dataset_name)

    def test_add_existing_component(self):
        component_id = "existing_component"
        dataset_name = "test_dataset"
        self.data_manager[dataset_name] = self.data
        self.data_manager._add_component(component_id, dataset_name)
        with pytest.raises(ValueError):
            self.data_manager._add_component(component_id, dataset_name)

    def test_get_component_data_nonexistent(self):
        dataset_name = "test_dataset"
        nonexistent_component = "nonexistent_component"
        self.data_manager[dataset_name] = self.data
        with pytest.raises(KeyError):
            self.data_manager._get_component_data(nonexistent_component)



"""Helper functions for models inside form folder."""
from typing import Union

from vizro._constants import ALL_OPTION
from vizro.models.types import MultiValueType, OptionsType, SingleValueType


def get_options_and_default(options: OptionsType, multi: bool = False):
    """Gets list of full options and default value based on user input type of `options`."""
    if multi:
        if all(isinstance(option, dict) for option in options):
            options = [{"label": ALL_OPTION, "value": ALL_OPTION}, *options]
        else:
            options = [ALL_OPTION, *options]

    if all(isinstance(option, dict) for option in options):
        default_value = options[0]["value"]  # type: ignore[index]
    else:
        default_value = options[0]

    return options, default_value


# Utils for validators
def is_value_contained(value: Union[SingleValueType, MultiValueType], options: OptionsType):
    """Checks if value is contained in a list."""
    if isinstance(value, list):
        return all(item in options for item in value)
    else:
        return value in options


# Validators for re-use
def validate_options_dict(cls, values):
    """Reusable validator for the "options" argument of categorical selectors."""
    if "options" not in values or not isinstance(values["options"], list):
        return values

    for entry in values["options"]:
        if isinstance(entry, dict) and not set(entry.keys()) == {"label", "value"}:
            raise ValueError("Invalid argument `options` passed. Expected a dict with keys `label` and `value`.")
    return values


def validate_value(cls, value, values):
    """Reusable validator for the "value" argument of categorical selectors."""
    if "options" not in values or not values["options"]:
        return value

    possible_values = (
        [entry["value"] for entry in values["options"]] if isinstance(values["options"][0], dict) else values["options"]
    )

    if value and not is_value_contained(value, possible_values):
        raise ValueError("Please provide a valid value from `options`.")

    return value


def validate_max(cls, max, values):
    """Reusable validator for the "max" argument for sliders."""
    if max is None:
        return max

    if values["min"] is not None and max < values["min"]:
        raise ValueError("Maximum value of slider is required to be larger than minimum value.")
    return max


def validate_slider_value(cls, value, values):
    """Reusable validator for the "value" argument for sliders."""
    if value is None:
        return value

    lvalue, hvalue = (value[0], value[1]) if isinstance(value, list) else (value, value)

    if (values["min"] is not None and not lvalue >= values["min"]) or (
        values["max"] is not None and not hvalue <= values["max"]
    ):
        raise ValueError("Please provide a valid value between the min and max value.")

    return value


def validate_step(cls, step, values):
    """Reusable validator for the "step" argument for sliders."""
    if step is None:
        return step

    if values["max"] is not None and step > (values["max"] - values["min"]):
        raise ValueError(
            "The step value of the slider must be less than or equal to the difference between max and min."
        )
    return step


def set_default_marks(cls, v, values):
    return v if values.get("step") is None else {}



from typing import List, Literal, Optional, Union

import pytest
from pydantic import Field, ValidationError
from typing_extensions import Annotated

import vizro.models as vm


class ChildX(vm.VizroBaseModel):
    type: Literal["child_x"] = "child_x"


class ChildY(vm.VizroBaseModel):
    type: Literal["child_y"] = "child_y"


class ChildZ(vm.VizroBaseModel):
    type: Literal["child_Z"] = "child_Z"


# ChildType does not include ChildZ initially.
ChildType = Annotated[Union[ChildX, ChildY], Field(discriminator="type")]


# These parent classes must be done as fixtures so that each test gets a fresh, unmodified copy of the class.
@pytest.fixture()
def Parent():
    # e.g. Parameter.selector: SelectorType
    class _Parent(vm.VizroBaseModel):
        child: ChildType

    return _Parent


@pytest.fixture()
def ParentWithOptional():
    # e.g. Filter.selector: Optional[SelectorType]
    class _ParentWithOptional(vm.VizroBaseModel):
        child: Optional[ChildType]

    return _ParentWithOptional


@pytest.fixture()
def ParentWithList():
    # e.g. Page.controls: List[ControlType] and Page.components: List[ComponentType]
    class _ParentWithList(vm.VizroBaseModel):
        child: List[ChildType]

    return _ParentWithList


@pytest.fixture()
def ParentWithForwardRef():
    class _ParentWithForwardRef(vm.VizroBaseModel):
        child: Annotated[Union["ChildXForwardRef", "ChildYForwardRef"], Field(discriminator="type")]  # noqa: F821

    _ParentWithForwardRef.update_forward_refs(ChildXForwardRef=ChildX, ChildYForwardRef=ChildY)
    return _ParentWithForwardRef


@pytest.fixture()
def ParentWithNonDiscriminatedUnion():
    class _ParentWithNonDiscriminatedUnion(vm.VizroBaseModel):
        child: Union[ChildX, ChildY]

    return _ParentWithNonDiscriminatedUnion


class TestDiscriminatedUnion:
    def test_no_type_match(self, Parent):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            Parent(child=child)

    def test_add_type_model_instantiation(self, Parent):
        Parent.add_type("child", ChildZ)
        parent = Parent(child=ChildZ())
        assert isinstance(parent.child, ChildZ)

    def test_add_type_dict_instantiation(self, Parent):
        Parent.add_type("child", ChildZ)
        parent = Parent(child={"type": "child_Z"})
        assert isinstance(parent.child, ChildZ)


class TestOptionalDiscriminatedUnion:
    # Optional[ChildType] does not work correctly as a discriminated union - pydantic turns it into a regular union.
    # Hence the validation error messages are not as expected. The tests of add_type pass because in practice a
    # discriminated union is not actually needed to achieve the desired behaviour. The union is still a regular one
    # even after add_type.
    @pytest.mark.xfail
    def test_no_type_match(self, ParentWithOptional):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            ParentWithOptional(child=child)

    # The current error message is the non-discriminated union one.
    def test_no_type_match_current_behaviour(self, ParentWithOptional):
        child = ChildZ()
        with pytest.raises(ValidationError, match="unexpected value; permitted: 'child_x'"):
            ParentWithOptional(child=child)

    def test_add_type_model_instantiation(self, ParentWithOptional):
        ParentWithOptional.add_type("child", ChildZ)
        parent = ParentWithOptional(child=ChildZ())
        assert isinstance(parent.child, ChildZ)

    def test_add_type_dict_instantiation(self, ParentWithOptional):
        ParentWithOptional.add_type("child", ChildZ)
        parent = ParentWithOptional(child={"type": "child_Z"})
        assert isinstance(parent.child, ChildZ)


class TestListDiscriminatedUnion:
    def test_no_type_match(self, ParentWithList):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            ParentWithList(child=[child])

    def test_add_type_model_instantiation(self, ParentWithList):
        ParentWithList.add_type("child", ChildZ)
        parent = ParentWithList(child=[ChildZ()])
        assert isinstance(parent.child[0], ChildZ)

    def test_add_type_dict_instantiation(self, ParentWithList):
        ParentWithList.add_type("child", ChildZ)
        parent = ParentWithList(child=[{"type": "child_Z"}])
        assert isinstance(parent.child[0], ChildZ)


class TestForwardRefDiscriminatedUnion:
    def test_no_type_match(self, ParentWithForwardRef):
        child = ChildZ()
        with pytest.raises(ValidationError, match="No match for discriminator 'type' and value 'child_Z'"):
            ParentWithForwardRef(child=child)

    def test_add_type_model_instantiation(self, ParentWithForwardRef, mocker):
        # Make it as if these are in vizro.models so that update_forward_refs call in add_type works on them.
        mocker.patch.dict(vm.__dict__, {"ChildXForwardRef": ChildX, "ChildYForwardRef": ChildY})
        ParentWithForwardRef.add_type("child", ChildZ)
        parent = ParentWithForwardRef(child=ChildZ())
        assert isinstance(parent.child, ChildZ)

    def test_add_type_dict_instantiation(self, ParentWithForwardRef, mocker):
        # Make it as if these are in vizro.models so that update_forward_refs call in add_type works on them.
        mocker.patch.dict(vm.__dict__, {"ChildXForwardRef": ChildX, "ChildYForwardRef": ChildY})
        ParentWithForwardRef.add_type("child", ChildZ)
        parent = ParentWithForwardRef(child={"type": "child_Z"})
        assert isinstance(parent.child, ChildZ)


def test_no_type_match(ParentWithNonDiscriminatedUnion):
    with pytest.raises(ValueError, match="Field 'child' must be a discriminated union"):
        ParentWithNonDiscriminatedUnion.add_type("child", ChildZ)



"""Script to generate JSON schema. For more information, run `hatch run schema --help`."""
import argparse
import json
import subprocess
import sys
from pathlib import Path

from vizro import __version__
from vizro.models import Dashboard

parser = argparse.ArgumentParser(description="Generate JSON schema.")
parser.add_argument("--check", help="check schema is up to date", action="store_true")
args = parser.parse_args()

# TODO: need to setup alias function to keep lowercase if wanted
schema_json = Dashboard.schema_json(indent=4, by_alias=False)
schema_path = Path(__file__).with_name(f"{__version__}.json")

if args.check:
    if json.loads(schema_path.read_text()) != json.loads(schema_json):
        sys.exit("JSON schema is out of date. Run `hatch run schema` to update it.")
    print("JSON schema is up to date.")  # noqa: T201
else:
    schema_path.write_text(schema_json)
    subprocess.run(f"pre-commit run prettier --files {schema_path}", shell=True, stdout=subprocess.DEVNULL)  # nosec



"""Unit tests for vizro.models.Button."""
import json

import dash_bootstrap_components as dbc
import plotly
import pytest
from dash import html

import vizro.models as vm
from vizro.actions import export_data


@pytest.fixture
def expected_button():
    return html.Div(
        [
            dbc.Button(
                id="button_id",
                children="Click me!",
                className="button_primary",
            ),
        ],
        className="button_container",
        id="button_id_outer",
    )


class TestButtonInstantiation:
    """Tests model instantiation and the validators run at that time."""

    def test_create_default_button(self):
        button = vm.Button()
        assert hasattr(button, "id")
        assert button.type == "button"
        assert button.text == "Click me!"
        assert button.actions == []

    @pytest.mark.parametrize(
        "text",
        ["Test", 123, 1.23, True, """# Header""", """<p>Hello </p>"""],
    )
    def test_create_button_with_optional(self, text):
        button = vm.Button(text=text)
        assert hasattr(button, "id")
        assert button.type == "button"
        assert button.text == str(text)
        assert button.actions == []

    def test_set_action_via_validator(self):
        button = vm.Button(actions=[vm.Action(function=export_data())])
        actions_chain = button.actions[0]
        assert actions_chain.trigger.component_property == "n_clicks"


class TestBuildMethod:
    def test_button_build(self, expected_button):
        button = vm.Button(id="button_id", text="Click me!").build()
        result = json.loads(json.dumps(button, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_button, cls=plotly.utils.PlotlyJSONEncoder))

        assert result == expected



"""General themed plotly template."""

from plotly import graph_objects as go


def create_template_common():
    """Create general themed plotly template.

    Returns:
    -------
        A plotly template object containing the general theme
        with optional keys specified here:
        https://plotly.com/python/reference/layout/
    """
    primary_font = "Inter, sans-serif, Arial"

    template_common = go.layout.Template()

    template_common.layout = go.Layout(
        # LAYOUT
        font_family=primary_font,
        font_size=14,
        title_font_family=primary_font,
        title_font_size=20,
        title_xref="container",
        title_yref="container",
        title_x=0,
        title_y=1,
        title_xanchor="left",
        title_yanchor="top",
        title_pad_l=80,
        title_pad_r=0,
        title_pad_t=24,
        title_pad_b=0,
        showlegend=True,
        legend_font_family=primary_font,
        legend_font_size=14,
        legend_orientation="h",
        legend_y=-0.20,
        legend_title_font_family=primary_font,
        legend_title_font_size=14,
        legend_title_text="",
        legend_bgcolor="rgba(0,0,0,0)",
        margin_l=80,
        margin_r=12,
        margin_t=64,
        margin_b=64,
        margin_pad=0,
        margin_autoexpand=True,
        coloraxis_colorbar_outlinewidth=0,
        coloraxis_colorbar_thickness=20,
        coloraxis_colorbar_showticklabels=True,
        coloraxis_colorbar_ticks="outside",
        coloraxis_colorbar_tickwidth=1,
        coloraxis_colorbar_ticklen=8,
        coloraxis_colorbar_tickfont_family=primary_font,
        coloraxis_colorbar_tickfont_size=14,
        coloraxis_colorbar_ticklabelposition="outside",
        coloraxis_colorbar_title_font_family=primary_font,
        coloraxis_colorbar_title_font_size=14,
        bargroupgap=0.1,
        uniformtext_minsize=12,
        uniformtext_mode="hide",
        # X AXIS
        xaxis_visible=True,
        xaxis_title_font_family=primary_font,
        xaxis_title_font_size=16,
        xaxis_title_standoff=8,
        xaxis_ticklabelposition="outside",
        xaxis_ticks="outside",
        xaxis_ticklen=8,
        xaxis_tickwidth=1,
        xaxis_showticklabels=True,
        xaxis_automargin=True,
        xaxis_tickfont_family=primary_font,
        xaxis_tickfont_size=14,
        xaxis_showline=True,
        xaxis_layer="below traces",
        xaxis_linewidth=1,
        xaxis_zeroline=False,
        # Y AXIS
        yaxis_visible=True,
        yaxis_title_font_family=primary_font,
        yaxis_title_font_size=16,
        yaxis_title_standoff=8,
        yaxis_ticklabelposition="outside",
        yaxis_ticks="outside",
        yaxis_ticklen=8,
        yaxis_tickwidth=1,
        yaxis_showticklabels=True,
        yaxis_automargin=True,
        yaxis_tickfont_family=primary_font,
        yaxis_tickfont_size=14,
        yaxis_showline=False,
        yaxis_layer="below traces",
        yaxis_linewidth=1,
        yaxis_zeroline=False,
    )

    return template_common
